#!/usr/bin/perl -w
#
# File     : fci-client.pl
# Author   : Bas Couwenberg <sebastic@xs4all.nl>
# License  : GPL
#

use strict;
use LWP::UserAgent;
use HTTP::Request::Common;
use File::Basename;
use File::Copy;
use Date::Calc;
use Sys::Hostname;
use Getopt::Mixed "nextOption";
use XML::Simple;

use constant ReturnImmediately => 0x10;
use constant ForwardOnly       => 0x20;

$|=1;

# configuration defaults
my $version  = '1.5.1';
my $hostname = hostname;
if($hostname =~ s/^(.*?)\..*$/$1/){} # strip domain name from hostname
my $fci_dir  = set_fci_dir();
my $bin_dir  = '/usr/local/bin/';
my $data_dir = '/usr/local/share/fci/';

my %cfg = (
		'self'                => 'fci-client.pl',
		'version'             => $version,
		
		'client-name'         => $hostname,
		'fah-dir'             => '/home/folding/',
		'fci-dir'             => $fci_dir,
		'bin-dir'             => $bin_dir,
				
		'queue-file'          => '/home/folding/queue.dat',
		'client-cfg-file'     => '/home/folding/client.cfg',
		'fahlog-file'         => '/home/folding/FAHlog.txt',
		'fahlog-prev-file'    => '/home/folding/FAHlog-Prev.txt',
		'myfolding-file'      => '/home/folding/MyFolding.html',
		'unitinfo-file'       => '/home/folding/unitinfo.txt',
		'zyx-file'            => '/home/folding/work/current.xyz',
		
		'fci-conf-file'       => $fci_dir.'fci-client.conf',
		'qdinfo-file'         => $fci_dir.'qdinfo.dat',
		'qd-bin'              => $bin_dir.'qd',
		'master-qd-bin'       => $bin_dir.'qd',
		'master-qdinfo-file'  => $data_dir.'qdinfo.dat',
		'qd'                  => 'qd',
		'no-qd-update'        => 0,
		'update-qd'           => 0,

		'lsb-release-bin'     => '/bin/lsb_release',
		'lspci-bin'           => '/usr/bin/lspci',
		'pciconf-bin'         => '/usr/sbin/pciconf',
		'system-profiler-bin' => '/usr/sbin/system_profiler',
		'nvidia-settings-bin' => '/usr/bin/nvidia-settings',
		'wmi-host'            => 'localhost',
		
		'url' 	              => 'http://example.com/fci/index.pl',
		'useragent'           => 'fci-client/'.$version,
		'username'            => '',
		'password'            => '',
		
		'save-config'         => 0, 
		'operatingsystem'     => $^O,
		'ds'                  => '/',
		'binarysafe'          => 0,
		'verbose'             => 0,
		'help'                => 0
	);

# set Operating System dependant (default) values,
# Linux is assumed
my %RegHash;
if($cfg{operatingsystem} eq 'linux') {
	$cfg{'lsb-release-bin'}     = find_binary('lsb_release', $cfg{'lsb-release-bin'});
	$cfg{'lspci-bin'}           = find_binary('lspci', $cfg{'lspci-bin'});
	$cfg{'nvidia-settings-bin'} = find_binary('nvidia-settings', $cfg{'nvidia-settings-bin'});
}
elsif($cfg{operatingsystem} eq 'cygwin') {
	# Cygwin (on Windows)
	
	require POSIX;
	import POSIX;
	
	require Win32;
	import Win32;

	require Win32::OLE;
	import Win32::OLE qw(in with);

	require Win32::TieRegistry;
	import Win32::TieRegistry (TiedHash => \%RegHash, Delimiter => '/');

	$cfg{'master-qd-bin'}      = '/usr/local/bin/qd.exe';
	$cfg{'master-qdinfo-file'} = '/usr/local/share/fci/qdinfo.dat';
	$cfg{'qd-bin'}             = '/usr/local/bin/qd.exe';
	$cfg{'qd'}                 = 'qd.exe';
	$cfg{'ds'}                 = '/'; # directory separator is /
	$cfg{'binarysafe'}         = 1;
}
elsif($cfg{operatingsystem} eq 'MSWin32') {
	# Windows
	
	require POSIX;
	import POSIX;
	
	require Win32;
	import Win32;

	require Win32::OLE;
	import Win32::OLE qw(in with);

	require Win32::TieRegistry;
	import Win32::TieRegistry (TiedHash => \%RegHash, Delimiter => '/');
			
	$cfg{'fah-dir'}          = 'C:\\Program Files\\Folding@Home';
	$cfg{'queue-file'}       = $cfg{'fah-dir'}.'\\queue.dat';
	$cfg{'client-cfg-file'}  = $cfg{'fah-dir'}.'\\client.cfg';
	$cfg{'fahlog-file'}      = $cfg{'fah-dir'}.'\\FAHlog.txt';
	$cfg{'fahlog-prev-file'} = $cfg{'fah-dir'}.'\\FAHlog-Prev.txt';
	$cfg{'myfolding-file'}   = $cfg{'fah-dir'}.'\\MyFolding.html';
	$cfg{'unitinfo-file'}    = $cfg{'fah-dir'}.'\\unitinfo.txt';
	$cfg{'xyz-file'}         = $cfg{'fah-dir'}.'\\work\\current.xyz';
	
	$cfg{'fci-conf-file'}    = $cfg{'fci-dir'} . '\\fci-client.conf';
	$cfg{'qd-bin'}           = $cfg{'fci-dir'} . '\\qd.exe';
	$cfg{'qdinfo-file'}      = $cfg{'fci-dir'} . '\\qdinfo.dat';
	$cfg{'qd'}	         = 'qd.exe';
	
	$cfg{'ds'}               = '\\'; # directory separator is \
	$cfg{'binarysafe'}       = 1;
}
elsif($cfg{operatingsystem} eq 'freebsd') {
	# FreeBSD
	
	$cfg{'master-qd-bin'}      = '/usr/local/bin/qd';
	$cfg{'master-qdinfo-file'} = '/usr/local/share/fci/qdinfo.dat';
	$cfg{'qd-bin'}             = '/usr/local/bin/qd';
	$cfg{'qd'}                 = 'qd.freebsd';
	$cfg{'pciconf-bin'}        = find_binary('pciconf', $cfg{'pciconf-bin'});
}
elsif($cfg{operatingsystem} eq 'openbsd') {
	# OpenBSD 

	$cfg{'master-qd-bin'}      = '/usr/local/bin/qd';
	$cfg{'master-qdinfo-file'} = '/usr/local/share/fci/qdinfo.dat';
	$cfg{'qd-bin'}             = '/usr/local/bin/qd';
	$cfg{'qd'}                 = 'qd.openbsd';
}
elsif($cfg{operatingsystem} eq 'darwin') {
	# MacOSX 
	# (not officially supported yet, but planned for the future)
	
	$cfg{'qd-bin'}        = '/usr/local/bin/qd.darwin';
	$cfg{'qd'}            = 'qd.darwin';
}

# global hash of files to upload if present
my %uploadable_file = (
			'client-cfg-file'  => 1,
			'fahlog-file'      => 1,
			'fahlog-prev-file' => 1,
			'myfolding-file'   => 1,
			'unitinfo-file'    => 1,
			'xyz-file'         => 1
		      );

my %skip_file;  # global hash of files not upload even if present
my %http_info;	# global hash of HTTP return codes for failed requests
my %parameter;  # global hash of set commandline parameters
my %from_conf;  # global hash of set configuration parameters

# get command line parameters
Getopt::Mixed::init("url=s
                     name=s
		     dir=s
		     fci-dir=s
		     useragent=s
		     username=s
		     password=s
		     client>name
		     skip-fahlog
		     skip-fahlog-prev
		     skip-client-cfg
		     skip-myfolding
		     skip-unitinfo
		     skip-xyz
		     qd=s
		     qdinfo=s
		     update-qd
		     no-qd-update
		     save-config
		     config-file=s
		     verbose
		     help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'url') {
		$cfg{'url'}                = $value;
		$parameter{'url'}          = 1;
	}
	elsif($option eq 'name') {
		$cfg{'client-name'}        = $value;
		$parameter{'client-name'}  = 1;
	}
	elsif($option eq 'dir') {
		$cfg{'fah-dir'}            = $value;
		$parameter{'fah-dir'}      = 1;
	}
	elsif($option eq 'fci-dir') {
		$cfg{'fci-dir'}            = $value;
		$parameter{'fci-dir'}      = 1;
	}
	elsif($option eq 'useragent') {
		$cfg{'useragent'}          = $value;
		$parameter{'useragent'}    = 1;
	}
	elsif($option eq 'username') {
		$cfg{'username'}           = $value;
		$parameter{'username'}     = 1;
	}
	elsif($option eq 'password') {
		$cfg{'password'}           = $value;
		$parameter{'password'}     = 1;
	}
	elsif($option eq 'qd') {
		$cfg{'qd-bin'}             = $value;
		$parameter{'qd-bin'}       = 1;
	}
	elsif($option eq 'qdinfo') {
		$cfg{'qdinfo-file'}        = $value;
		$parameter{'qdinfo-file'}  = 1;
	}
	elsif($option eq 'update-qd') {
		$cfg{'update-qd'}          = 1;
		$parameter{'update-qd'}    = 1;
	}
	elsif($option eq 'no-qd-update') {
		$cfg{'no-qd-update'}       = 1;
		$parameter{'no-qd-update'} = 1;
	}
	elsif($option eq 'save-config') {
		$cfg{'save-config'}        = 1;
	}
	elsif($option eq 'config-file') {
		$cfg{'fci-conf-file'}      = $value;
	}
	elsif($option eq 'verbose') {
		$cfg{'verbose'}            = 1;
		$parameter{$option}        = 1;
	}
	elsif($option eq 'help') {
		$cfg{'help'}               = 1;
	}

	if($option =~ /^skip-(.*)$/) {
		$skip_file{$1.'-file'} = "Commandline Override";
		$parameter{$option}    = 1;
	}
}
Getopt::Mixed::cleanup();


# load the configuration from file into the config hash
if(-r $cfg{'fci-conf-file'}) {
	if(%cfg = load_config(%cfg)) {
		print "Successfully loaded configuration file.\n" if($cfg{verbose});
	}
	else {
		print "Error: Failed to load configuration file!\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});
}

# print usage and exit if --help
if($cfg{help} == 1) {
	print_usage();
	exit;
}	


my $command_line  = $0;
   $command_line .= " --url $cfg{'url'}" 		if($parameter{'url'}              || $from_conf{'url'});
   $command_line .= " --username $cfg{'username'}" 	if($parameter{'username'}         || $from_conf{'username'});
   $command_line .= " --password ********"		if($parameter{'password'}         || $from_conf{'password'});
   $command_line .= " --useragent $cfg{'useragent'}"	if($parameter{'useragent'}        || $from_conf{'useragent'});
   $command_line .= " --name $cfg{'client-name'}"	if($parameter{'client-name'}      || $from_conf{'client-name'});
   $command_line .= " --dir $cfg{'fah-dir'}"		if($parameter{'fah-dir'}          || $from_conf{'fah-dir'});
   $command_line .= " --skip-xyz"			if($parameter{'skip-xyz'}         || $from_conf{'skip-xyz'});
   $command_line .= " --skip-client-cfg"		if($parameter{'skip-client-cfg'}  || $from_conf{'skip-client-cfg'});
   $command_line .= " --skip-unitinfo"			if($parameter{'skip-unitinfo'}    || $from_conf{'skip-unitinfo'});
   $command_line .= " --skip-fahlog"			if($parameter{'skip-fahlog'}      || $from_conf{'skip-fahlog'});
   $command_line .= " --skip-fahlog-prev"		if($parameter{'skip-fahlog-prev'} || $from_conf{'skip-fahlog-prev'});
   $command_line .= " --skip-myfolding"			if($parameter{'skip-myfolding'}   || $from_conf{'skip-myfolding'});
   $command_line .= " --fci-dir $cfg{'fci-dir'}"	if($parameter{'fci-dir'}          || $from_conf{'fci-dir'});
   $command_line .= " --verbose"			if($parameter{'verbose'}          || $from_conf{'verbose'});

print "Effective command line:\n $command_line\n\n" if($cfg{verbose});


# suffix the directory separator to directory paths
$cfg{'fci-dir'} .= $cfg{ds} if(substr($cfg{'fci-dir'}, -1, 1) ne $cfg{ds});
$cfg{'fah-dir'} .= $cfg{ds} if(substr($cfg{'fah-dir'}, -1, 1) ne $cfg{ds});

# check the URL to the FCI server index.pl file
if($cfg{'url'} =~ m!/index.pl$!) {
	# all is fine
}
# suffix index.pl to the upload URL if the URL points to a directory
elsif($cfg{'url'} =~ m!/$!) {
	$cfg{'url'} .= 'index.pl';
}
else {
	# let the user fix it
	print "Invalid URL: $cfg{'url'} (Missing index.pl location)\n";
	exit;
}

## update paths with user defined directory
$cfg{'queue-file'}       = $cfg{'fah-dir'}.'queue.dat';
$cfg{'client-cfg-file'}  = $cfg{'fah-dir'}.'client.cfg';
$cfg{'fahlog-file'}      = $cfg{'fah-dir'}.'FAHlog.txt';
$cfg{'fahlog-prev-file'} = $cfg{'fah-dir'}.'FAHlog-Prev.txt';
$cfg{'myfolding-file'}   = $cfg{'fah-dir'}.'MyFolding.html';
$cfg{'unitinfo-file'}    = $cfg{'fah-dir'}.'unitinfo.txt';
$cfg{'xyz-file'}         = $cfg{'fah-dir'}.'work'.$cfg{ds}.'current.xyz';


# check the paths of the required programs,
# update their paths if it's not the preset
%cfg = find_qd_files(%cfg);

# verify the client name
if(!$cfg{'client-name'} || $cfg{'client-name'} !~ /^[0-9a-zA-Z\-\_]+$/) {
	print "Error: Invalid client name! ($cfg{'client-name'})\n";
	if($cfg{'client-name'}) {
		print "Allowed characters: 0-9 a-z A-Z - _\n";
	}
	exit;
}


# verify fci directory and its content
#
# The fci directory ~/.fci/ will be created if it doesn't exist, and the
# files inside it will be verified if the directory does exists.
# 
# The files inside ~/.fci/ are the qd binary and its data file qdinfo.dat.
# These two files can be updated by fci-client.pl by downloading their latest
# versions from the FCI server.
if(!verify_fci_dir()) {
	print "Error: Failed to verify the fci directory! ($cfg{'fci-dir'})\n";
	exit;
}
print "\n" if($cfg{verbose});


# update the paths to the qd binary and qdinfo.dat file
#
# We prefer to use the qd files in ~/.fci/ instead of /usr/ because we have 
# write permissions to the ones under our home directory qdinfo.dat will be 
# automatically updated with the latest version from the FCI server, and the qd
# binary can be optionally updated using the --update-qd commandline parameter

%cfg = update_qd_paths(%cfg);


# save the configuration to file if requested
if($cfg{'save-config'}) {
	#print "save-config: this functionality is currently disabled.\n";
	if(save_config(%cfg)) {
		print "Configuration file saved.\n" if($cfg{verbose});
	}
	else {
		print "Error: Could not save configuration file! ($cfg{'fci-conf-file'})\n";
	}
	exit;
}


# verify permissions to required files
if(!valid_permissions()) {
	print "Error: Incorrect permissions to required files!\n";
	exit;
}
print "\n" if($cfg{verbose});


# create useragent object which will handle the HTTP connections
my $ua  = LWP::UserAgent->new();
   $ua->agent($cfg{useragent});
   
# verify access to the server
print "Upload URL: $cfg{url}\nVerifying access to URL ... " if($cfg{verbose});

my $return_status = valid_url($cfg{url});
if($return_status >= 1) {
	if($return_status == 2) {
		# authentication required,
		
		print "Contacted!\n" if($cfg{verbose});
		
		if($cfg{verbose}) {
			print "The server responded with a HTTP 401 status code.\n";
		}
		($ua, $return_status) = set_http_auth($ua);
		
		# exit if there was an error while settings the auth creds
		exit if(!$return_status);
	}
	else {
		# status == 1, HTTP 200 OK
		# everything is fine, and no auth required
		
		print "Success!\n" if($cfg{verbose});
	}
}
else {
	print "FAILED!\n" if($cfg{verbose});
	
	print "Error: Server did not respond properly!\n";
	print "HTTP Status: $http_info{$cfg{url}}{status}\n";
	
	exit;
}
print "\n" if($cfg{verbose});


# download only qdinfo.dat by default,
# only update the qd binary if explicitly requested with --update-qd
# skip all updates if requested with --no-qd-update
if(!$cfg{'no-qd-update'}) {
	if(!update_qd_files($ua, $cfg{url}, $cfg{'update-qd'})) {
		print "Error: Failed to update qd files!\n";
	}
}
else {
	print "Skipping updating of qd files.\n" if($cfg{verbose});
}
print "\n" if($cfg{verbose});


# upload the files to the server
if(!upload_files($ua, $cfg{url})) {
	print "Error: Failed to upload files!\n";
}


################################[ Subroutines ]################################

##############
#
#     Routine: get_date()
#  Parameters: optionally a hash with the keys date and/or time set to something
#     Returns: the date in yyyy-mm-dd format, if only date is set
#              the time in hh:mm:ss format, if only time is set
#              the full date in yyyy-mm-dd hh:mm:ss format, if neither date and
#              time are set
#              both the date and the time, if they're both set
# Description: Get the contents and other info of uploaded data
#

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: iso_date()
#  Parameters: date in format: Tue Jan 17 13:37:47 2006
#     Returns: date in format: 2006-01-17 13:37:47
# Description: Convert a date in qd format to ISO format
#

sub iso_date {
        my $date     = $_[0];
        my $iso_date = $date;

        # Tue Jan 17 13:37:47 2006
        if($date =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)$/) {
                my $dow   = $1;
                my $month = $2;
                my $day   = $3;
                my $time  = $4;
                my $year  = $5;

                $month = Date::Calc::Decode_Month($month);

                $day   = '0'.$day   if(length($day)   == 1);
                $month = '0'.$month if(length($month) == 1);

                $iso_date = "$year-$month-$day $time";
        }

        # 2006-01-17 13:37:47
        return $iso_date;
}


##############
#
#     Routine: normalize_filesize()
#  Parameters: filesize
#              current size unit
#     Returns: normalized filesize and unit name
# Description: convert a filesize to its smallest unit
#     Example: 2097152 KB -> 2 GB
#

sub normalize_filesize {
        my $size = $_[0] =~ /^\d+$/        ? $_[0] : 0;
        my $unit = $_[1] =~ /^[KMGTPE]?B$/ ? $_[1] : 'B';

        my $u     = 0;
        my @units = qw/B KB MB GB TB PB EB/;
        foreach(@units) {
                if($unit =~ /^$_/i) {
                        if($size >= 1024) {
                                $size /= 1024;
                                $unit  = $units[$u+1];
                        }
                        else {
                                last;
                        }
                }
                $u++;
        }

        return ($size, $unit);
}


##############
#
#     Routine: time_format()
#  Parameters: time in seconds
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert time in seconds to readable format
#

sub time_format {
        my $value = shift;
        my $ret;

        if($value && ($value >= 0)) {
                # convert time in seconds to human readable format
                # e.g. 65 seconds == 1m5s
		
		if($value >= 60) {
                        my $years   = 0;
                        my $months  = 0;
                        my $days    = 0;
                        my $hours   = 0;
                        my $minutes = 0;
                        my $seconds = $value;

		        $minutes = int($seconds / 60);
		        $seconds = int($seconds - ($minutes * 60));

			if($minutes >= 60) {
				$hours   = int($minutes / 60);
			        $minutes = $minutes - ($hours * 60);

				if($hours >= 24) {
					$days  = int($hours / 24);
			        	$hours = $hours - ($days * 24);

					if($days >= 30.41667) {
						$months = int($days / 30.41667);
			        		$days   = $days - ($months * 30.41667);

						if($months >= 12) {
							$years  = int($months / 12);
			        			$months = $years - ($years * 12);
						}
					}
				}
			}

                        my $time  = '';
			   $time .= $years  ."y"   if($years   > 0);
                           $time .= $months ."m"   if($months  > 0);
                           $time .= $days   ."d"   if($days    > 0);
                           $time .= $hours  ."h"   if($hours   > 0);
                           $time .= $minutes."m"   if($minutes > 0);
                           $time .= $seconds."s"   if($seconds > 0);
                       
			$ret = $time;
		}
		else {
			my $seconds = $value;

			my $time = '';
			   $time .= $seconds."s"   if($seconds > 0);   
				
			$ret = $time;
		}
        }
        else {
		$ret = $value;
        }

        return $ret;
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file = $_[0];
	
	print "Reading file from disk: $file ... " if($cfg{verbose});
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($cfg{verbose});

		return $data;
	}
	else {
		print "FAILED!\n" if($cfg{verbose});
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode   => file premission in octal
# 	         binary => flag to set binary mode (only needed on Windows)
#     Returns: 1 for successfull save or
# 	       undef for unsuccessfull save
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	print "Saving file to disk: $filename ... " if($cfg{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($cfg{verbose});
		
		open(F, "> $filename") || return undef; 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($cfg{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($cfg{verbose});
			chmod oct($option{mode}), $filename;
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($cfg{verbose});
		return undef;
	}
}


##############
#
#     Routine: fetch_url()
#  Parameters: user-agent object
#              url
#     Returns: page content on success or
#              undef in case of no success
# Description: retrieve a file via an HTTP GET request
# 

sub fetch_url {
	my $ua  = $_[0];
	my $url = $_[1];
	
	my $req = HTTP::Request->new(GET => $url);
	
	print "Retrieving URL: $url ... " if($cfg{verbose});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{verbose});
		
		return $res->content;
	}
	else {
		print "Failed!\n" if($cfg{verbose});
		print "Error: Failed to retrieve URL! ($url)\n";
		print "HTTP Status: ".$res->code." ".$res->message."\n";
			
		return undef;
	}
}
																	

##############
#
#     Routine: valid_url()
#  Parameters: url
#     Returns: 1 on success or
#              2 in case of authentication required or
#              0 otherwise
# Description: verify the access to a file via a HTTP HEAD request and
#              detect requirement of HTTP Authentication
# 

sub valid_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	   $ua->agent($cfg{useragent});
	
	my $req = HTTP::Request->new(HEAD => $url);
	my $res = $ua->request($req);
	
	# store realm and http status info in global hash
	$http_info{$url}{realm}  = $res->header('WWW-Authenticate');
	$http_info{$url}{status} = $res->code." ".$res->message;

	if($res->is_success()) {
		return 1;
	}
	elsif($res->code == 401) {
		return 2;
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: set_http_auth()
#  Parameters: useragent object
#     Returns: useragent object and
#              1 on success,
#              0 on failure
# Description: update the useragent object with authentication credentials
#  

sub set_http_auth {
	my $ua = $_[0];
	
	if($cfg{password} eq '') {
		print "Authentication required, please supply password!\n";
		return($ua, 0);
		exit;
	}
	# If we have no username, assume client authentication
	$cfg{username} = 'client_'.$cfg{'client-name'} if(!$cfg{username});
		
	my $netloc   = $cfg{url};
	my $realm    = $http_info{$cfg{url}}{realm};
			
	if(!$realm) {
		print "Error: Cannot find realm name in HTTP response!\n";
		return($ua, 0);
	}
	
	# extract the realm name from the HTTP response string
	$realm =~ s/^.*?realm="(.*)"$/$1/;
	
	# extract netlocation from URL,
	# set default port if no port specified in URL
	$netloc  =~ s/^http:\/\/(.*?)\/.*$/$1/;
	$netloc .= ':80' if($netloc !~ /:\d+$/);
				
	if($cfg{verbose}) {
		print "Using authentication credentials:\n";
		print " Netlocation: $netloc\n";
		print " Realm      : $realm\n";
		print " Username   : $cfg{username}\n";
		print " Password   : ********\n";
	}
			
	# set credentials in UserAgent object
	$ua->credentials($netloc, $realm, $cfg{username}, $cfg{password});
	
	return($ua, 1);
}


##############
#
#     Routine: upload_files()
#  Parameters: useragent object
#              url
#     Returns: 1 on success,
#              0 on failure
# Description: upload files to the server for central monitoring
#  

sub upload_files {
	my $ua  = $_[0];
	my $url = $_[1];

	print "Running qd binary to parse the Folding\@Home client files.\n" if($cfg{verbose});
	my ($qd_data, $qd) = get_qd_data();
	
	# hash of upload fields
	my %content = ( 'submit'      => 1,
			'action'      => 'upload',
			'client-name' => $cfg{'client-name'},
			'client-os'   => $^O,
			'fci-version' => $cfg{version},
			'qd-data'     => $qd_data,
		      );
	
	# get a list of files which will be uploaded
	my @upload_files = get_upload_files($qd);

	# add files to upload to upload fields
	foreach(@upload_files) {
		$content{$_} = [ $cfg{$_} ];
	}

	# get cpu and ram info
	my ($cpu, $ram) = get_cpu_ram(%cfg);
	
	# add the cpu and ram info to upload if found
	if($cpu && $cpu ne '') {
		$content{'client-cpu'} = $cpu;
	}
	if($ram && $ram ne '') {
		$content{'client-ram'} = $ram;
	}

	# get GPU info
	my $gpu = get_gpu(%cfg);
	if($gpu && $gpu ne '') {
		$content{'client-gpu'} = $gpu;
	}

	# get Operating System name, version, hardware & uptime
	my %os = get_os_info(%cfg);
	if($os{name} && $os{name} ne '') {
		$content{'client-os-name'} = $os{name};		
	}
	if($os{version} && $os{version} ne '') {
		$content{'client-os-version'} = $os{version};		
	}
	if($os{hardware} && $os{hardware} ne '') {
		$content{'client-os-hardware'} = $os{hardware};		
	}
	if($os{uptime} && $os{uptime} ne '') {
		$content{'client-os-uptime'} = $os{uptime};		
	}
	
	# the request object
	my $req = POST $cfg{url},
	          Content_Type => 'form-data',
		  Content      => \%content;
	
	# execute the request
	print "\nUploading data to the server ... " if($cfg{verbose});
	my $res = $ua->request($req);
	
	if($res->is_success()) {
		if($cfg{verbose} == 1) {
			print "Success.\n\n";
		}
			
		if($res->content =~ /Error: /) {
			print "\nWarning: The server encounterd one or more errors while processing the upload!\n";

			if(!$cfg{verbose}) {
				print "Forcing verbose output to display the server respose.\n\n";
				$cfg{verbose} = 1;
			}
		}
			
		if($cfg{verbose}) {
			print "##############################[ Server Response ]##############################\n";
			print $res->content;
			print "\n";
			print "##############################[ Server Response ]##############################\n";
		}
	}
	elsif($res->code eq '401') {
		# Uploads require authentication
		
		$http_info{$cfg{url}}{realm}  = $res->header('WWW-Authenticate');
		$http_info{$cfg{url}}{status} = $res->code." ".$res->message;
		
		if($cfg{verbose}) {
			print "Authentication Required!\n";
			print "The server responded with a HTTP 401 status code.\n";
			print "Retrying with Authentication credentials.\n\n";
		}
		($ua, $return_status) = set_http_auth($ua, %cfg);
		
		# execute the request
		print "\nUploading data to the server ... " if($cfg{verbose});
		
		my $res = $ua->request($req);
		if($res->is_success()) {
			if($cfg{verbose} == 1) {
				print "Success.\n\n";
			}
		
			if($res->content =~ /Error: /) {
				print "\nWarning: The server encounterd one or more errors while processing the upload!\n";
	
				if(!$cfg{verbose}) {
					print "Forcing verbose output to display the server respose.\n\n";
					$cfg{verbose} = 1;
				}
			}

			if($cfg{verbose}) {
				print "##############################[ Server Response ]##############################\n";
				print $res->content;
				print "\n";
				print "##############################[ Server Response ]##############################\n";
			}
		}
		else {
			print "FAILED!\n" if($cfg{verbose});
		
			print "Error: Failed to upload data to: $cfg{url}\n";
			print "Server Response: ".$res->code.' '.$res->message."\n";
	
			return 0;
		}
	}
	else {
		print "FAILED!\n" if($cfg{verbose});
		
		print "Error: Failed to upload data to: $cfg{url}\n";
		print "Server Response: ".$res->code.' '.$res->message."\n";

		return 0;
	}

	return 1;
}


##############
#
#     Routine: update_qd_files()
#  Parameters: useragent object
#              url
#     Returns: 1 on success,
#              0 on failure
# Description: download the latest version of the qd files 
#              if newer versions are available
#  

sub update_qd_files {
	my $ua        = $_[0];
	my $url       = $_[1];
	my $update_qd = $_[2];

	my $qdinfo_url  = $url;
	   $qdinfo_url  =~ s/index.pl$/qd-data\/qdinfo.dat/;
	my $qdinfo_data = '';

	print "Updating qd files.\n" if($cfg{verbose});

	# download qdinfo.dat
	if(valid_url($qdinfo_url)) {
		if($qdinfo_data = fetch_url($ua, $qdinfo_url)) {
			# continue
		}
		else {
			return 0;
		}
	}
	else {
		print "Error: Invalid URL to qdinfo.dat! ($qdinfo_url)\n";
		return 0;
	}

	# check downloaded qdinfo.dat against local version
	# if the downloaded version is newer, save it to disk
	# if the downloaded version has a higher functional revision than the 
	#  local qd, also download the qd binary
	# 
	# the downloaded and local version of qdinfo.dat may be identical,
	# even when in reality they are not. The version string does not take
	# the time of update into consideration. The pg (point table generation
	# time) does take the time into consideration, but is not displayed by 
	# qd
	
	my $local_data      = read_file($cfg{'qdinfo-file'});
	my $local_version   = get_qdinfo_version($local_data);
	my $fetched_version = get_qdinfo_version($qdinfo_data);
	
	if($cfg{verbose}) {
		print "\n";
		print "  Local version: $local_version\n";
		print "Fetched version: $fetched_version\n\n";
	}
	
	my $local_fr   = get_qd_functionalrevision($cfg{'qd-bin'});
	my $fetched_fr = get_qdinfo_functionalrevision($qdinfo_data);
	
	if($cfg{verbose}) {
		print "  Local functional revision: $local_fr\n";
		print "Fetched functional revision: $fetched_fr\n\n";
	}

	my $local_pg   = get_qdinfo_generationtime($local_data);
	my $fetched_pg = get_qdinfo_generationtime($qdinfo_data);

	if($cfg{verbose}) {
		my $local_date   = scalar localtime $local_pg;
		my $fetched_date = scalar localtime $fetched_pg;

		print "  Local generation time: $local_date\n";
		print "Fetched generation time: $fetched_date\n\n";
	}


	# strip the - chars from the date strings 
	# to mathematically compare the dates
	$local_version   =~ s/-//g;
	$fetched_version =~ s/-//g;
	
	# compare version
	if($fetched_version > $local_version || $fetched_pg > $local_pg) {
		print "Fetched version is more recent.\n" if($cfg{verbose});
		
		if(save_file($cfg{'qdinfo-file'}, $qdinfo_data)) {
			# continue
		}
		else {
			print "Error: Could not save file! ($cfg{'qdinfo-file'})\n";
			return 0;
		}
	}
	elsif($fetched_version < $local_version) {
		print "Fetched version is less recent, keeping local version of qdinfo.dat.\n" if($cfg{verbose});
	}
	else {
		print "Versions are the same, no update of qdinfo.dat necessary.\n" if($cfg{verbose});
	}
		
	# only download the qd binary if explicitly requested
	if($update_qd) {
		# compare functional revision
		if($fetched_fr > $local_fr) {
			print "\nFetched functional revision is more recent.\n" if($cfg{verbose});
	
			# download latest qd binary
			my $qd_url  = $url;
			   $qd_url  =~ s/index.pl$/qd-data\/$cfg{qd}/;
			my $qd_data = '';
			
			if(valid_url($qd_url)) {
				if($qd_data = fetch_url($ua, $qd_url)) {
					# continue
				}
				else {
					return 0;
				}
				
				# save file to disk
				if(save_file($cfg{'qd-bin'}, $qd_data, binary => $cfg{binarysafe})) {
					# continue
				}
				else {
					print "Error: Could not save file! ($cfg{'qd-bin'})\n";
					return 0;
				}
			}
			else {
				print "Error: Invalid URL to qd binary! ($qd_url)\n";
				return 0;
			}
		}
		elsif($fetched_fr < $local_fr) {
			print "Fetched functional revision is less recent, keeping local version of qd.\n" if($cfg{verbose});
		}
		else {
			print "Functional revisions are the same, no update of qd necessary.\n" if($cfg{verbose});
		}
	}
	else {
		if($cfg{verbose}) {
			print "Not updating qd binary, use --update-qd to also update the binary.\n";
			print "The binary will only be updated if there is a more recent functional revision.\n";
		}
	}
	
	print "Finished updating qd files.\n" if($cfg{verbose});
	return 1;
}


##############
#
#     Routine: get_qdinfo_version()
#  Parameters: qdinfo.dat content
#     Returns: ISO 8601 date string
# Description: find the date stamp in qdinfo.dat and convert to ISO 8601
# 

sub get_qdinfo_version {
	my $data = $_[0];
	
	# extract the date from the content
	# da 24 May 2005 (rph)
	if($data =~ /da (\d+) (\w+) (\d+) \(.*?\)/) {
		my ($day, $month, $year) = ($1, $2, $3);

		$month = Date::Calc::Decode_Month($month);

		$day   = '0'.$day   if(length($day)   == 1);
		$month = '0'.$month if(length($month) == 1);
		
		return "$year-$month-$day";
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: get_qdinfo_functionalrevision()
#  Parameters: qdinfo.dat content
#     Returns: functional revision
# Description: find the functional revision in qdinfo.dat
# 

sub get_qdinfo_functionalrevision {
	my $data = $_[0];

	# extract the functional revision from the content
	# fr 029
	if($data =~ /fr (\d+)/) {
		return $1;
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: get_qd_functionalrevision()
#  Parameters: path to qd binary
#     Returns: functional revision
# Description: find the functional revision of a qd binary
# 

sub get_qd_functionalrevision {
	my $qd_bin = $_[0];
	
	if(-x $qd_bin) {
		# execute the binary with -v (version) parameter
		my $qd_data = `"$qd_bin" -v`;

		# qd released 24 May 2005 (fr 031)
		if($qd_data =~ /qd released \d+ \w+ \d+ \(fr (\d+)\)/) {
			return $1;
		}
		else {
			return 0;
		}
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: get_qdinfo_generationtime()
#  Parameters: qdinfo.dat content
#     Returns: point table generation time as integer (epoch 2000)
# Description: find the point table generation time in qdinfo.dat
# 

sub get_qdinfo_generationtime {
	my $data = $_[0];

	# extract the point table generation time from the content
	# pg 1145E762
	if($data =~ /pg (\w+)/) {
		return (hex($1) + 946684800);
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: get_qd_data()
#  Parameters: none
#     Returns: output of the qd command
#              under in case of an error
# Description: run the qd binary to parse queue.dat in the folding@Home client directory
# 

sub get_qd_data {
	# ~/.fci/qd -n ~/.fci/qdinfo.dat -f ~/software/folding

	my $fah_dir = $cfg{'fah-dir'};
	
	# strip trailing \ from path 
	# if running on Windows and it is appended to the path
	#
	# this fixed an error qd might throw:
	#  Can't open <C:\Program Files\Folding"\queue.dat> file
	if($^O eq 'MSWin32' && $fah_dir =~ s/\\$//){}
	
	print "Executing: \"$cfg{'qd-bin'}\" -n \"$cfg{'qdinfo-file'}\" -f \"$fah_dir\" -x\n" if($cfg{verbose});
	my $qd_data = `"$cfg{'qd-bin'}" -n "$cfg{'qdinfo-file'}" -f "$fah_dir" -x`;

	# Verify qd output
        my $qd = XMLin( $qd_data,
		        KeyAttr    => { index => 'number', server => 'number' },
                        ForceArray => [ 'index', 'server' ]
                      ) || print "Warning: Failed to parse qd data: $@\n";
        
	#print Dumper($qd);
	if($qd->{error}) {
		print "Warning: Queue dump error: $qd->{error}\n";
	}

	return ($qd_data, $qd);
}


##############
#
#     Routine: get_upload_files()
#  Parameters: none
#     Returns: list of key names of files to upload
# Description: create a list of files to upload based on their availability and 
#              user settings
# 

sub get_upload_files {
	my $qd = shift;

	my @file = ();

	# loop over the files to skip, and remove them from the upload list
	print "\nComposing list of files to upload:\n" if($cfg{verbose});
	foreach my $key (keys %uploadable_file) {
		# disable the entry, if told to skip it
		if($skip_file{$key}) {
			$uploadable_file{$key} = 0;
		}
		else {
			# Don't upload the current.xyz, if it likely doesn't belong to the current work unit
			if($key eq 'xyz-file') {
				if(!exists $qd->{error} && 
				   exists $qd->{queue}->{current_index} &&
				   exists $qd->{queue}->{index}->{$qd->{queue}->{current_index}}->{status} &&
				   $qd->{queue}->{index}->{$qd->{queue}->{current_index}}->{status} eq 'folding now'
				) {
					my $min_date = $qd->{queue}->{index}->{$qd->{queue}->{current_index}}->{begin_timestamp};
					my $max_date = get_date();
					my $xyz_date = (lstat($cfg{$key}))[9]; # mtime

					$min_date = iso_date($min_date);
					$xyz_date = iso_date(scalar localtime $xyz_date);

					$min_date =~ s/[- :]//g;
					$max_date =~ s/[- :]//g;
					$xyz_date =~ s/[- :]//g;
				
					if(!($xyz_date >= $min_date && $xyz_date <= $max_date)) { 
						$uploadable_file{$key} = 0;

						$skip_file{$key} = "Not Current";
					}
				}
				else {
					$uploadable_file{$key} = 0;
					
					$skip_file{$key} = "Unverifiable";
				}
			}
			# Don't upload the unitinfo.txt, if it is too big to upload
			elsif($key eq 'unitinfo-file') {
				my $size = (lstat($cfg{$key}))[7]; # size

				#    most seen normal filesize: 156 & 157 bytes
				# largest seen normal filesize: 191       bytes
				if($size > 512) {
					$uploadable_file{$key} = 0;
					
					$skip_file{$key} = "Filesize Limit";
				}
			}
		}
	}

	# check permission of upload files
	foreach my $key (sort keys %uploadable_file) {
		print " Checking: $cfg{$key} ... " if($cfg{verbose});

		if($uploadable_file{$key}) {
			if(-r $cfg{$key}) {
				print "Upload.\n" if($cfg{verbose});
				push @file, $key;		
			}
			else {
				print "Skip. (Cannot read file)\n" if($cfg{verbose});
			}
		}
		else {
			if($cfg{verbose}) {
				print "Skip.";
				print " ($skip_file{$key})" if($skip_file{$key});
				print "\n"
			}
		}
	}
	
	return @file;
}


##############
#
#     Routine: get_cpu_ram()
#  Parameters: %cfg
#     Returns: CPU speed in Mhz and amount of RAM in MB
#              0 in case of an error
# Description: find the CPU speed and amount of installed RAM of through the Operating System
# 

sub get_cpu_ram {
	my %cfg = @_;

	print "\nLooking up CPU and RAM information:\n" if($cfg{verbose});
	
	if($cfg{operatingsystem} eq 'MSWin32' || $cfg{operatingsystem} eq 'cygwin') {
		# Windows
		
		my $cpu = '';
		my $ram = '';
			
		# Connect to WMI service
		my $WMI = Win32::OLE->GetObject("winmgmts:\\\\$cfg{'wmi-host'}\\root\\CIMV2") || die "Error: WMI connection failed to: $cfg{'wmi-host'}\n";

		my $win32_computersystem = $WMI->ExecQuery("SELECT * FROM Win32_ComputerSystem", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_computersystem)
		{
			$ram = $obj->{TotalPhysicalMemory} .' B';

			last;
		}

		my $win32_processor = $WMI->ExecQuery("SELECT * FROM Win32_Processor", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_processor)
		{
			$cpu  = $obj->{Name};
			$cpu .= ' '. $obj->{MaxClockSpeed} .' MHz @' 		if($obj->{MaxClockSpeed} ne $obj->{CurrentClockSpeed});
			$cpu .= ' '. $obj->{CurrentClockSpeed} .' MHz';
			$cpu .= ' '. $obj->{L2CacheSize} .' KB L2 Cache'        if($obj->{L2CacheSize} > 0);
			$cpu  =~ s/^\s*//;

			last;
		}

		if($cfg{verbose}) {
			print " CPU: $cpu\n";
			print " RAM: $ram\n";
		}

		return ($cpu, $ram);
	}
	elsif($cfg{operatingsystem} eq 'freebsd') {
		# FreeBSD
		
		my $file = '/var/run/dmesg.boot';
		if(-r $file) {
			my $cpu = '';
			my $ram = '';
			
			my $cpu_line = `grep "CPU:" "$file"`;
			my $origin   = `grep "Origin" "$file"`;
			my $ram_line = `grep "real mem" "$file"`;
			
			#  Origin = "AuthenticAMD"  Id = 0x671  Stepping = 1
			#if($origin =~ /Origin = "(.*?)"/) {
			#	$cpu = $1;
			#}

			#CPU: AMD Duron(tm)  (1300.06-MHz 686-class CPU)
			if($cpu_line =~ /CPU: (.*?)\s+\((.*?)-MHz/) {
				$cpu .= " " if($cpu ne '');
				$cpu .= "$1 $2 MHz";
			}

			#real memory  = 536084480 (511 MB)
			if($ram_line =~ /real memory\s+= (\d*)/) {
				$ram = $1;
				$ram = ($ram / 1024) . " kB" if($ram ne '');
			}

			if($cfg{verbose}) {
				print " CPU: $cpu\n";
				print " RAM: $ram\n";
			}
			
			return ($cpu, $ram);
		}
		else {
			return 0;
		}
	}
	elsif($cfg{operatingsystem} eq 'openbsd') {
		# OpenBSD 
		
		my $file = '/var/run/dmesg.boot';
		if(-r $file) {
			my $cpu = '';
			my $ram = '';
			
			my $cpu_line = `grep -i ^cpu "$file"`;
			my $ram_line = `grep "real mem" "$file"`;
			
			#cpu0: AMD Athlon(TM) XP 2400+ ("AuthenticAMD" 686-class) 2.01 GHz
			if($cpu_line =~ /^cpu\d: (.* \wHz)/) {
				$cpu = $1;
			}

			#real mem  = 267952128 (261672K)
			if($ram_line =~ /real mem\s+= (\d*)/) {
				$ram = $1;
				$ram = ($ram / 1024) . " kB" if($ram ne '');
			}

			if($cfg{verbose}) {
				print " CPU: $cpu\n";
				print " RAM: $ram\n";
			}
			
			return ($cpu, $ram);
		}
		else {
			return 0;
		}
	}
	elsif($cfg{operatingsystem} eq 'darwin') {
		# Mac OS X/PPC
		#
		# $ system_profiler SPHardwareDataType
		# Hardware:
		#
		#    Hardware Overview:
		#
		#      Machine Name: Power Mac G4
		#      Machine Model: PowerMac3,6
		#      CPU Type: PowerPC G4  (3.3)
		#      Number Of CPUs: 2
		#      CPU Speed: 1.42 GHz
		#      L2 Cache (per CPU): 256 KB
		#      L3 Cache (per CPU): 2 MB
		#      Memory: 2 GB
		#      Bus Speed: 167 MHz
		#      Boot ROM Version: 4.6.0f1
		#      Serial Number: XB3110TMN1Y
		#
		# Mac OS X/Intel
		#
		# $ system_profiler SPHardwareDataType
		# Hardware:
		#
		#    Hardware Overview:
		#
		#      Machine Name: Mac mini
		#      Machine Model: Macmini1,1
		#      Processor Name: Intel Core Duo
		#      Processor Speed: 1.83 GHz
		#      Number Of Processors: 1
		#      Total Number Of Cores: 2
		#      L2 Cache (per processor): 2 MB
		#      Memory: 1 GB
		#      Bus Speed: 667 MHz
		#      Boot ROM Version: MM11.0055.B08
		#      SMC Version: 1.3f4
		#      Serial Number: YM6374AMW0B
		#
		
		# OS X v10.2 - v10.x: /usr/sbin/system_profiler
		if(-x $cfg{'system-profiler-bin'}) {
			my $output = `$cfg{'system-profiler-bin'} SPHardwareDataType`;

			my $cpu = '';
			my $ram = '';
			
			# CPU Speed: 1.42 GHz
			if($output =~ /CPU Speed: ((\d+\.\d+|\d+) \wHz)/) {
				$cpu = $1;
			}
			# Processor Speed: 1.83 GHz
			elsif($output =~ /Processor Speed: ((\d+\.\d+|\d+) \wHz)/) {
				$cpu = $1;
			}

			# CPU Type: PowerPC G4  (3.3)
			if($output =~ /CPU Type: (.*)/) {
				$cpu = $1 ." ". $cpu;
			}
			# Processor Name: Intel Core Duo
			elsif($output =~ /Processor Name: (.*)/) {
				$cpu = $1 ." ". $cpu;
			}
			
			# Memory: 2 GB
			if($output =~ /Memory: ((\d+\.\d+|\d+) \wB)/) {
				$ram = $1;
			}
			
			if($cfg{verbose}) {
				print " CPU: $cpu\n";
				print " RAM: $ram\n";
			}
			
			return ($cpu, $ram);
		}
		# OS X v10.0 - v10.1: AppleSystemProfiler
		elsif(-x "/usr/sbin/AppleSystemProfiler" && -x "/usr/sbin/sysctl") {
			my $cpu = '';
			my $ram = '';
			
			my $output = `/usr/sbin/AppleSystemProfiler`;
			# Machine speed = 666 MHz
			if($output =~ /Machine speed = ((\d+|\d+\.\d+) \wHz)/) {
				$cpu = $1;
			}
			
			#sysctl -n hw.physmem
			$output = `/usr/sbin/sysctl -n hw.physmem`;
			if($output =~ /^(\d+)$/) {
				$ram = $1 . " B";
			}
			
			if($cfg{verbose}) {
				print " CPU: $cpu\n";
				print " RAM: $ram\n";
			}
			
			return ($cpu, $ram);
		}
		else {
			return 0;
		}
	}
	elsif($cfg{operatingsystem} eq 'linux') {
		# Linux
		if(-r "/proc/cpuinfo" && -r "/proc/meminfo") {
			my $cpu_vendor = '';
			my $cpu_model  = '';
			my $cpu_mhz    = '';
			my $cpu_cache  = '';
			my $mem_total  = '';
			
			# kernel file of cpu information
			open(F, "/proc/cpuinfo") || return 0;
			while(<F>) {
				chomp;
				
				#vendor_id       : AuthenticAMD
				if(/^vendor_id\s+: (.*)$/i) {
					#$cpu_vendor = $1;
				}
				#model name      : AMD Sempron(TM) 2200+
				elsif(/^model name\s+: (.*)$/i) {
					$cpu_model = $1;
				}
				#cpu MHz         : 1499.906
				elsif(/^cpu MHz\s+: (.*)$/i) {
					$cpu_mhz = $1;
				}
				#cache size      : 256 KB
				elsif(/^cache size\s+: (.*)$/i) {
					$cpu_cache = $1;
				}
			}
			close F;
			
			# kernel file of memory information
			open(F, "/proc/meminfo") || return 0;
			while(<F>) {
				chomp;
				
				#MemTotal:      1816828 kB
				if(/MemTotal:\s+(.*)$/i) {
					$mem_total = $1;
				}
			}
			close F;

			my $cpu  = '';
			   $cpu .= "$cpu_vendor "     if($cpu_vendor ne '');
			   $cpu .= "$cpu_model "      if($cpu_model  ne '');
			   $cpu .= "$cpu_mhz MHz "    if($cpu_mhz    ne '');
			   $cpu .= "$cpu_cache Cache" if($cpu_cache  ne '');
			my $ram  = $mem_total;
			
			if($cfg{verbose}) {
				print " CPU: $cpu\n";
				print " RAM: $ram\n";
			}

			return ($cpu, $ram);
		}
	}
	else {
		# unsupported Operating System
		print "Warning: Unsupported Operating System: $^O\n" if($cfg{verbose});
		return;
	}
}


##############
#
#     Routine: get_gpu()
#  Parameters: %cfg
#     Returns: GPU information
#              0 in case of an error
# Description: find the GPU information for the card installed
# 

sub get_gpu {
	my %cfg = @_;

	print "\nLooking up GPU information:\n" if($cfg{verbose});
	
	if($cfg{operatingsystem} eq 'MSWin32' || $cfg{operatingsystem} eq 'cygwin') {
		# Windows
		
		my $gpu = '';
			
		# Connect to WMI service
		my $WMI = Win32::OLE->GetObject("winmgmts:\\\\$cfg{'wmi-host'}\\root\\CIMV2") || die "Error: WMI connection failed to: $cfg{'wmi-host'}\n";

		my $win32_videocontroller = $WMI->ExecQuery("SELECT * FROM Win32_VideoController", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_videocontroller)
		{
			my ($size, $unit) = ($obj->{AdapterRAM}, 'B');
			   ($size, $unit) = normalize_filesize($size, $unit);
			    $size = sprintf("%.2f", $size);

			$gpu  = $obj->{Caption};
			$gpu .= ' '. $size .' '. $unit	 		if($obj->{AdapterRAM});
			$gpu .= ' ('. $obj->{DriverVersion} .')'	if($obj->{DriverVersion});

			last;
		}

		if($cfg{verbose}) {
			print " GPU: $gpu\n";
		}

		return $gpu;
	}
	elsif($cfg{operatingsystem} eq 'freebsd') {
		# FreeBSD
		
		my $gpu = '';

		if(-x $cfg{'pciconf-bin'}) {
			my @output = `$cfg{'pciconf-bin'} -lv`;
			my $parse  = 0;
			foreach(@output) {
				# vgapci0@pci0:0:15:0:    class=0x030000 card=0x040515ad chip=0x040515ad rev=0x00 hdr=0x00
				#     vendor     = 'VMware Inc.'
				#     device     = 'VGA 4.0.5 Virtual SVGA'
				#     class      = display
				#     subclass   = VGA
				if(!$parse && /^vga\w+@\w+:.*?:\s+/) {
					$parse = 1;	
				}
				elsif($parse && /^\s+vendor\s+= '(.*?)'/) {
					$gpu = $1;
				}
				elsif($parse && /^\s+device\s+= '(.*?)'/) {
					$gpu .= ' ' if($gpu ne '');
					$gpu .= $1;
					last;
				}
				elsif($parse && !/^\s+\w+/) {
					$parse = 0;
				}
			}
		}
		else {
			print "Warning: Cannot execute pciconf binary: $cfg{'pciconf-bin'}\n";
			return 0;
		}

		if($cfg{verbose}) {
			print " GPU: $gpu\n";
		}
	
		return $gpu;
	}
	elsif($cfg{operatingsystem} eq 'openbsd') {
		# OpenBSD 
		
		my $gpu = '';
		   $gpu = 'Warning: GPU detection not supported on OpenBSD';

		if($cfg{verbose}) {
			print " GPU: $gpu\n";
		}
	
		return 0;
	}
	elsif($cfg{operatingsystem} eq 'darwin') {
		# Mac OS X/PPC
		
		my $gpu = '';

                # OS X v10.2 - v10.x: /usr/sbin/system_profiler
                if(-x "/usr/sbin/system_profiler") {
			# $ system_profiler SPDisplaysDataType
			# Graphics/Displays:
			# 
			#     Intel GMA 950:
			# 
			#       Chipset Model: GMA 950
			#       Type: Display
			#       Bus: Built-In
			#       VRAM (Total): 64 MB of shared system memory
			#       Vendor: Intel (0x8086)
			#       Device ID: 0x27a2
			#       Revision ID: 0x0003
			#       Displays:
			#         Display Connector:

                        my @output = `/usr/sbin/system_profiler SPDisplaysDataType`;

			foreach(@output) {
				#       Chipset Model: GMA 950
				if(/Chipset Model: (.*?)\s*$/) {
					$gpu = $1;
				}
				#       VRAM (Total): 64 MB of shared system memory
				elsif(/VRAM \(Total\): (.*?)\s*$/) {
					$gpu .= ' ' if($gpu ne '');	
					$gpu .= $1;
				}
				#       Vendor: Intel (0x8086)
				elsif(/Vendor: (.*?) \(0x\d+\)\s*$/) {
					if($gpu ne '') {
						$gpu = $1 .' '. $gpu 
					}
					else {
						$gpu = $1;
					}
				}
			}
		}
		else {
			print "Warning: Cannot execute system_profiler binary: system_profiler\n";
			return 0;
		}

		if($cfg{verbose}) {
			print " GPU: $gpu\n";
		}
	
		return $gpu;
	}
	elsif($cfg{operatingsystem} eq 'linux') {
		# Linux
		
		my $gpu = '';

		if(-x $cfg{'lspci-bin'}) {
			my @output = `$cfg{'lspci-bin'}`;
			foreach(@output) {
				# 01:00.0 VGA compatible controller: nVidia Corporation NV11 [GeForce2 MX/MX 400] (rev b2)
				if(/^(.*?) VGA compatible controller: (.*?)\s*$/) {
					my $pci_id = $1;
					my $name   = $2;

					$gpu = $2;
					last;
				}
			}

			if($gpu =~ /(nVidia)/i && -x $cfg{'nvidia-settings-bin'}) {
				# use nvidia-settings to query the device properties
				
				my $gpu_vendor = $1;

				my $output = `$cfg{'nvidia-settings-bin'} -ctrl-display=:0 -q gpus -q fans 2>&1`;
				#
				#1 GPU on anubis2:0
				#
				#    [0] anubis2:0[gpu:0] (GeForce 9500 GT)
				#
				if($output =~ /^\s+\[(\d+)\]\s+(.*?)(\[gpu:\d+\])\s+\((.*?)\)/m) {
					my $gpu_num  = $1;
					my $xdisplay = $2;
					my $gpu_id   = $3;
					my $gpu_name = $4;

					$gpu = $gpu_vendor .' '. $gpu_name;

					if(my $vram = `$cfg{'nvidia-settings-bin'} -ctrl-display=:0 -q ${gpu_id}/VideoRam -t`) {
						chomp($vram);

						my ($size, $unit) = ($vram, 'KB');
						   ($size, $unit) = normalize_filesize($size, $unit);
						    $size = sprintf("%.2f", $size);
		
						$gpu .= ' '. $size .' '. $unit;
					}

					if(my $driver_version = `$cfg{'nvidia-settings-bin'} -ctrl-display=:0 -q ${gpu_id}/NvidiaDriverVersion -t`) {
						chomp($driver_version);

						$gpu .= ' ('. $driver_version .')';
					}

					if(my $core_temp = `$cfg{'nvidia-settings-bin'} -ctrl-display=:0 -q ${gpu_id}/GPUCoreTemp -t`) {
						chomp($core_temp);

						$gpu .= ' Temp: '. $core_temp .'C';
					}
					
					# 
					# 1 Fan on osiris:0
					# 
					#     [0] osiris:0[fan:0] (Fan 0)
					# 
					if($output =~ /^\s+\[(\d+)\]\s+(.*?)(\[fan:\d+\])\s+\((.*?)\)/m) {
						my $fan_num  = $1;
						my $xdisplay = $2;
						my $fan_id   = $3;
						my $fan_name = $4;
				
						if(my $fan_speed = `$cfg{'nvidia-settings-bin'} -ctrl-display=:0 -q ${fan_id}/GPUCurrentFanSpeed -t`) {
							chomp($fan_speed);
	
							$gpu .= ' Fan: '. $fan_speed .'%';
						}
					}
				}
			}
		}
		else {
			print "Warning: Cannot execute lspci binary: $cfg{'lspci-bin'}\n";
			return 0;
		}

		if($cfg{verbose}) {
			print " GPU: $gpu\n";
		}
	
		return $gpu;
	}
	else {
		# unsupported Operating System
		print "Warning: Unsupported Operating System: $^O\n" if($cfg{verbose});
		return;
	}
}


##############
#
#     Routine: get_os_info()
#  Parameters: %cfg
#     Returns: CPU speed in Mhz and amount of RAM in MB
#              0 in case of an error
# Description: find out various bits of info on the OS
# 

sub get_os_info {
	my %cfg = @_;
	
	print "\nLooking up Operating System information.\n" if($cfg{verbose});
	
	my %info;
	if($cfg{operatingsystem} eq 'MSWin32' || $cfg{operatingsystem} eq 'cygwin') {
		# Windows

		#     name: Microsoft(R) Windows(R) XP Professional x64 Edition
		#  version: 5.2.3790 Service Pack 2 Build 3790
		# hardware: x64-based PC
		#   uptime: 1 Days, 15 Hours, 32 Minutes, 9 Seconds

		# Connect to WMI service
		my $WMI = Win32::OLE->GetObject("winmgmts:\\\\$cfg{'wmi-host'}\\root\\CIMV2") || die "Error: WMI connection failed to: $cfg{'wmi-host'}\n";

		my $win32_operatingsystem = $WMI->ExecQuery("SELECT * FROM Win32_OperatingSystem", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_operatingsystem)
		{
		        my $date_boot = $obj->{LastBootUpTime};
		        my $date_now  = $obj->{LocalDateTime};
		
		        # 20080716122003.085000+060
		        if($date_boot =~ /^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})\.(\d{6})(\+|\-)(\d{3})$/) {
		                my $year        = $1;
		                my $month       = $2;
		                my $day         = $3;
		                my $hour        = $4;
		                my $minute      = $5;
		                my $second      = $6;
		                my $millisecond = $7;
		                my $sign        = $8;
		                my $tz_min_diff = $9;
		
		                $date_boot = "$year-$month-$day $hour:$minute:$second";
		        }
		
		        if($date_now =~ /^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})\.(\d{6})(\+|\-)(\d{3})$/) {
		                my $year        = $1;
		                my $month       = $2;
		                my $day         = $3;
		                my $hour        = $4;
		                my $minute      = $5;
		                my $second      = $6;
		                my $millisecond = $7;
		                my $sign        = $8;
		                my $tz_min_diff = $9;
		
		                $date_now = "$year-$month-$day $hour:$minute:$second";
		        }
		
		        my @boot = split /-|:| /, $date_boot;
		        my @now  = split /-|:| /, $date_now;
		
		        my $year1  = $boot[0];
		        my $month1 = $boot[1];
		        my $day1   = $boot[2];
		        my $hour1  = $boot[3];
		        my $min1   = $boot[4];
		        my $sec1   = $boot[5];
		
		        my $year2  = $now[0];
		        my $month2 = $now[1];
		        my $day2   = $now[2];
		        my $hour2  = $now[3];
		        my $min2   = $now[4];
		        my $sec2   = $now[5];
		
		        my ($D_y,$D_m,$D_d, $Dh,$Dm,$Ds) = Date::Calc::Delta_YMDHMS($year1,$month1,$day1, $hour1,$min1,$sec1,
		                                                                    $year2,$month2,$day2, $hour2,$min2,$sec2);
		
		        my $uptime = "";
		        if($D_y > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$D_y years";
		        }
		        if($D_m > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$D_m months";
		        }
		        if($D_d > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$D_d days";
		        }
		        if($Dh > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$Dh hours";
		        }
		        if($Dm > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$Dm minutes";
		        }
		        if($Ds  > 0) {
		                $uptime .= ", " if($uptime ne "");
		                $uptime .= "$Ds seconds";
		        }

			$info{name}     = $obj->{Caption};

			$info{version}  = $obj->{Version};
			$info{version} .= ' '. $obj->{CSDVersion}		if($obj->{CSDVersion});
			$info{version} .= ' Build '. $obj->{BuildNumber}	if($obj->{BuildNumber});

			$info{uptime}   = $uptime				if($uptime);

			last;
		}
		
		my $win32_processor = $WMI->ExecQuery("SELECT * FROM Win32_Processor", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_processor)
		{
			$info{uptime} .= ', '	if($info{uptime});
			$info{uptime} .= 'load: '. $obj->{LoadPercentage} .'%';

			last;
		}

		my $win32_computersystem = $WMI->ExecQuery("SELECT * FROM Win32_ComputerSystem", "WQL", ReturnImmediately | ForwardOnly);

		foreach my $obj (in $win32_computersystem)
		{
			$info{hardware} = $obj->{SystemType};

			last;
		}
		
		if($cfg{operatingsystem} eq 'cygwin') {
			# Cygwin, has uname :-)
	
			# -a: CYGWIN_NT-5.1 osiris-winxp 1.5.18(0.132/4/2) 2005-07-02 20:30 i686 unknown unknown Cygwin
			#
			# -s: CYGWIN_NT-5.1
			# -n: osiris-winxp
			# -r: 1.5.18(0.132/4/2)
			# -v: 2005-07-02 20:30
			# -m: i686
			# -p: unknown
			# -i: unknown
			# -o: Cygwin
	
			my $uname = `uname`;
			if($uname) {
				my $machine_hardware = `uname -m`;
			        my $nodename         = `uname -n`;
			        my $os_release       = `uname -r`;
			        my $os_kernel        = `uname -s`;
			        my $os_version       = `uname -v`;
			        my $processor        = `uname -p`;
			        my $kernel_ident     = `uname -i`;
				my $os_name          = `uname -o`;
			}
			
			# Cygwin, has uname :-)
		
			# -a: CYGWIN_NT-5.1 osiris-winxp 1.5.18(0.132/4/2) 2005-07-02 20:30 i686 unknown unknown Cygwin
			#
			# -s: CYGWIN_NT-5.1
			# -n: osiris-winxp
			# -r: 1.5.18(0.132/4/2)
			# -v: 2005-07-02 20:30
			# -m: i686
			# -p: unknown
			# -i: unknown
			# -o: Cygwin
	
			#my $uname = `uname`;
			#if($uname) {
			#	my $machine_hardware = `uname -m`;
			#        my $nodename         = `uname -n`;
			#        my $os_release       = `uname -r`;
			#        my $os_kernel        = `uname -s`;
			#        my $os_version       = `uname -v`;
			#        my $processor        = `uname -p`;
			#        my $kernel_ident     = `uname -i`;
			#	my $os_name          = `uname -o`;
			#
			#	foreach(qw/m n r s v p i o/) {
			#		print "uname -$_: ".`uname -$_`;	
			#	}
			#}
		}
		
		my ($sysname, $nodename, $release, $version, $machine) = POSIX::uname();
		
		my $ticks   = Win32::GetTickCount();
		my $seconds = $ticks / 1000;
		my $uptime  = time_format($seconds);

		#print "Sysname : $sysname\n";
		#print "Nodename: $nodename\n";
		#print "Release : $release\n";
		#print "Version : $version\n";
		#print "Machine : $machine\n";
		#print "Uptime  : $uptime\n";
		if(!$info{name} || $info{name} =~ /^\s*$/) {
			$info{name} = $sysname;
		}
		if(!$info{version} || $info{version} =~ /^\s*$/) {
			$info{version} = $release .' '.$version;
		}
		if(!$info{hardware} || $info{hardware} =~ /^\s*$/) {
			$info{hardware} = $machine;
		}
		if(!$info{uptime} || $info{uptime} =~ /^\s*$/) {
			$info{uptime} = $uptime;
		}
	}
	elsif($cfg{operatingsystem} eq 'freebsd') {
		# FreeBSD
		
		# uname switches:
		#
		# -a FreeBSD localhost.localdomain 6.0-RELEASE FreeBSD 6.0-RELEASE #0: 
		#    Thu Nov  3 09:36:13 UTC 2005     root@x64.samsco.home:/usr/obj/usr/src/sys/GENERIC  i386
		#
		# -s FreeBSD
		# -n localhost.localdomain
		# -r 6.0-RELEASE
		# -v FreeBSD 6.0-RELEASE #0: Thu Nov  3 09:36:13 UTC 2005     
		#    root@x64.samsco.home:/usr/obj/usr/src/sys/GENERIC 
		# -m i386
		# 
		# -i GENERIC
		# -p i386
		
		#chomp(my $nodename         = `uname -n`;
		#chomp(my $os_version       = `uname -v`;
		
		# uname -p is not part of uname -a
		#chomp(my $processor        = `uname -p`);

		# uname -i is not part of uname -a
		#chomp(my $kernel_ident     = `uname -i`);
		
		chomp(my $os_kernel        = `uname -s`);
		chomp(my $os_release       = `uname -r`);
		chomp(my $machine_hardware = `uname -m`);
		chomp(my $uptime           = `uptime`);

		$info{name}     = $os_kernel;
		$info{version}  = $os_release;
		$info{hardware} = $machine_hardware;
		$info{uptime}   = $uptime;
	}
	elsif($cfg{operatingsystem} eq 'openbsd') {
		# OpenBSD 
	
		# uname switches:
		#
		# -a OpenBSD localhost.localdomain 3.8 GENERIC#138 i386
		# 
		# -s OpenBSD
		# -n localhost.localdomain
		# -r 3.8
		# -v GENERIC#138
		# -m i386
		#
		# -p AMD Athlon(TM) XP 2400+ ("AuthenticAMD" 686-class, 256KB L2 cache)
		
		#chomp(my $nodename         = `uname -n`);
		#chomp(my $os_version       = `uname -v`);
		
		# uname -p is not part of uname -a
		#chomp(my $processor        = `uname -p`);
		
		chomp(my $os_kernel        = `uname -s`);
		chomp(my $os_release       = `uname -r`);
		chomp(my $machine_hardware = `uname -m`);
		chomp(my $uptime           = `uptime`);
		
		$info{name}     = $os_kernel;
		$info{version}  = $os_release;
		$info{hardware} = $machine_hardware;
		$info{uptime}   = $uptime;
	}
	elsif($cfg{operatingsystem} eq 'darwin') {
		# MacOSX 
		
		# uname switches:
		#
		# -a Darwin hostname.example.com 8.7.0 Darwin Kernel Version 8.7.0: Fri May 26 15:20:53 PDT 2006; 
		#    root:xnu-792.6.76.obj~1/RELEASE_PPC Power Macintosh powerpc
		#
		# -s Darwin
		# -n hostname.example.com
		# -r 8.7.0
		# -v Darwin Kernel Version 8.7.0: Fri May 26 15:20:53 PDT 2006; 
		#    root:xnu-792.6.76.obj~1/RELEASE_PPC
		# -m Power Macintosh
		# -p powerpc
		
		#chomp(my $nodename         = `uname -n`;
		#chomp(my $os_version       = `uname -v`;
		#chomp(my $processor        = `uname -p`);

		chomp(my $os_kernel        = `uname -s`);
		chomp(my $os_release       = `uname -r`);
		chomp(my $machine_hardware = `uname -m`);
		chomp(my $uptime           = `uptime`);

		$info{name}     = $os_kernel;
		$info{version}  = $os_release;
		$info{hardware} = $machine_hardware;
		$info{uptime}   = $uptime;

		# OS X v10.2 - v10.x: /usr/sbin/system_profiler
		if(-x "/usr/sbin/system_profiler") {
			my $output = `/usr/sbin/system_profiler SPSoftwareDataType`;
			
			my $os_name = '';
			
			# Software:
			#
			#    System Software Overview:
			#
			#      System Version: Mac OS X 10.4.8 (8L2127)
			#      Kernel Version: Darwin 8.8.1
			#      Boot Volume: Macintosh HD
			#      Computer Name: joe-bar
			#      User Name: Joe MacUser (joe)
	
			if($output =~ /System Version: (.*)/) {
				$info{name} = $1;	
			}
			if($output =~ /Kernel Version: (.*)/) {
				$info{version} = $1;
			}
		}	
	}
	elsif($cfg{operatingsystem} eq 'linux') {
		# Linux
		
		# uname switches:
		#
		# -a Linux osiris 2.6.11-1-k7 #1 Mon Jun 20 21:26:23 MDT 2005 i686 GNU/Linux
		#
		# -s Linux
		# -n osiris
		# -r 2.6.11-1-k7
		# -v #1 Mon Jun 20 21:26:23 MDT 2005
		# -m i686
		# -o GNU/Linux
		
		#chomp(my $nodename         = `uname -n`);
		#chomp(my $os_version       = `uname -v`);
		#chomp(my $os_name          = `uname -o`);
		
		chomp(my $os_kernel        = `uname -s`);
		chomp(my $os_release       = `uname -r`);
		chomp(my $machine_hardware = `uname -m`);

		chomp(my $uptime           = `uptime`);
		#if(-r "/proc/uptime") {
		#	open(F, "/proc/uptime") || die "Cannot open file ($!)";
		#	while(<F>) {
		#		if(//) {
		#		}
		#	}
		#	close F;
		#}

		$info{name}     = $os_kernel;
		$info{version}  = $os_release;
		$info{hardware} = $machine_hardware;
		$info{uptime}   = $uptime;

		# Detect distro specific info
		if(-r "/etc/debian_version") {
			$info{name}    = "Debian GNU/Linux";
			$info{version} = "$os_kernel $os_release";

			my $version = read_file("/etc/debian_version");
			
			# 3.1
			if($version =~ /^(.*?)\n$/) {
				$info{name} .= " ".$1;
			}
		}
		elsif(-r "/etc/gentoo-release") {
			$info{name} = "Gentoo Linux";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/gentoo-release");
			
			# Gentoo Base System version 1.12.1
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/fedora-release") {
			$info{name} = "Fedora Core";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/fedora-release");
			
			# Fedora Core release 6 (Zod)
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/redhat-release") {
			$info{name} = "Red Hat";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/redhat-release");
			
			# Red Hat Enterprise Linux AS release 4 (Nahant Update 3)
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/SuSE-release") {
			$info{name} = "SUSE";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/SuSE-release");
			
			# openSUSE 10.2 (i586)
			# VERSION = 10.2
			if($version =~ /^(.*?)\nVERSION = (.*?)$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/arch-release") {
			$info{name} = "Arch Linux";
			$info{version} = "$os_kernel $os_release";
		}
		elsif(-r "/etc/slackware-version" || -r "/etc/slackware-release") {
			$info{name} = "Slackware";
			$info{version} = "$os_kernel $os_release";
				
			my $version = '';
			if(-r "/etc/slackware-version") {
				$version = read_file("/etc/slackware-version");
			}
			elsif(-r "/etc/slackware-release") {
				$version = read_file("/etc/slackware-release");
			}
			
			# Slackware release 8.1 (Midas)
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
			
			# SLAX is a derivative of Slackware
			if(-r "/etc/slax-version") {
				$version = read_file("/etc/slax-version");

				# SLAX 5.1.8
				if($version =~ /^(.*?)\n$/) {
					$info{name} = "$1 ($info{name})";
				}
			}
		}
		elsif(-r "/etc/mandrake-release") {
			$info{name} = "Mandrake";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/mandrake-release");
			
			# Mandrake Linux release 8.1 (Vitamin) for i586
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/UnitedLinux-release") {
			$info{name} = "UnitedLinux";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/UnitedLinux-release");
			
			# UnitedLinux 1.0 (ia64)
			# VERSION = 1.0
			if($version =~ /^(.*?)\nVERSION = (.*?)$/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/sun-release") {
			$info{name} = "Sun";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/sun-release");
			
			# Sun Java Desktop System, Release 2 -build 10b (GA)
			# Assembled 30 March 2004
			if($version =~ /^(.*?)\n/) {
				$info{name} = $1;
			}
		}
		elsif(-r "/etc/yellowdog-release") {
			$info{name} = "Yellow Dog Linux";
			$info{version} = "$os_kernel $os_release";
	
			my $version = read_file("/etc/yellowdog-release");
			
			# Yellow Dog Linux release 4.0 (Orion)
			if($version =~ /^(.*?)\n$/) {
				$info{name} = $1;
			}
		}

		# Detect Distro info via LSB
		if(-x $cfg{'lsb-release-bin'}) {
			my %lsb_release = get_lsb_release_data($cfg{'lsb-release-bin'});
			
			# Debian 3.1:   Debian
			# Kubuntu 6.10: Ubuntu
			# MEPIS 8.0:	MEPIS
			if($lsb_release{distrib_id} ne '') {
				$info{name} = $lsb_release{distrib_id};
			}
			# Debian 3.1:	Debian GNU/Linux
			# Debian sid:	Debian GNU/Linux unstable (sid)
			# Kubuntu 6.10:	Ubuntu 6.10
			# MEPIS 8.0:	MEPIS 8.0 (upgradable from Debian etch)
			if($lsb_release{distrib_description} ne '') {
				$info{name} = $lsb_release{distrib_description};
			}
			# Debian 3.1:	3.1
			# Debian sid:	unstable
			# Kubuntu 6.10:	6.10
			# MEPIS 8.0:	8.0
			if($lsb_release{distrib_release} ne '') {
				if($info{name} eq $lsb_release{distrib_id} ||
				   $info{name} !~ /$lsb_release{distrib_release}/
				) {
					$info{name} .= " ".$lsb_release{distrib_release};
				}
			}
			# Debian 3.1:	sarge
			# Debian sid:	sid
			# Kubuntu 6.10:	edgy
			# MEPIS 8.0:	lenny
			if($lsb_release{distrib_codename} ne '') {
				if($info{name} eq $lsb_release{distrib_id} ||
				   $info{name} eq "$lsb_release{distrib_id} $lsb_release{distrib_release}"
				) {
					$info{name} .= " (". $lsb_release{distrib_codename} .")";
				}
			}

			# MEPIS does not use the expected format for the LSB DISTRIB_DESCRIPTION:
			# DISTRIB_ID DISTRIB_RELEASE (DISTRIB_CODENAME)
			if($lsb_release{distrib_id}       eq 'MEPIS' && 
			   $lsb_release{distrib_release}  ne ''      &&
			   $lsb_release{distrib_codename} ne ''
			) {
				$info{name}  = $lsb_release{distrib_id} .' '. $lsb_release{distrib_release};
				$info{name} .= ' (Debian '. $lsb_release{distrib_codename} .')';
			}
		}
	}
	else {
		# unsupported Operating System
		print "Warning: Unsupported Operating System: $^O\n";
	}

	if($cfg{verbose}) {
		print " Name    : $info{name}\n";
		print " Version : $info{version}\n";
		print " Hardware: $info{hardware}\n";
		print " Uptime  : $info{uptime}\n";
	}

	return %info;
}


##############
#
#     Routine: update_qd_paths()
#  Parameters: %cfg
#     Returns: updated %cfg
# Description: check the preset paths for the required programs
# 

sub update_qd_paths {
	my %cfg = @_;

	# the qd binary used 
	
	print "Updating path of qd binary ... \n" if($cfg{verbose});

	# the qd binary used should be ~/.fci/qd, /usr/local/bin/qd or /usr/bin/qd, 
	# ./qd{.openbsd|.freebsd|.darwin} on UNIX systems
	# on Windows the qd binary used should be in the same directory as this file, fci-client.pl
	
	my $found = 0;
	my @paths;
	if($^O ne 'MSWin32' && $^O ne 'cygwin') {
		@paths = ($cfg{'fci-dir'}.'qd', '/usr/local/bin/qd', '/usr/bin/qd', dirname($0).'/'.$cfg{qd});
	}
	elsif($^O eq 'cygwin') {
		@paths = ($cfg{'fci-dir'}.'qd.exe', '/usr/local/bin/qd.exe', '/usr/bin/qd.exe', dirname($0).'/'.$cfg{qd});
	}
	else {
		@paths = (dirname($0).'\\qd.exe');
	}
		
	foreach(@paths) {
		print " $_\n" if($cfg{verbose});

		# exists and is executable?
		if(-e $_ && -x $_) {
			print "Using qd at: $_\n" if($cfg{verbose});
			$cfg{'qd-bin'} = $_;
			$found = 1;
			last;
		}
	}
		
	if(!$found) {
		print "Error: Cannot find qd binary! Use --qd to set the path.\n";
		exit;
	}
	print "\n" if($cfg{verbose});

	
	# qdinfo.dat file
	print "Updating path of qdinfo.dat file ... \n" if($cfg{verbose});
	
	# the qdinfo.dat file used should be ~/.fci/qdinfo.dat, /usr/local/share/fci/qdinfo.dat or 
	# /usr/share/fci/qdinfo.dat on UNIX systems
	# on Windows the qdinfo.dat file should be in the same directory as this file, fci-client.pl
	
	$found = 0;
	if($^O ne 'MSWin32') {
		@paths = ($cfg{'fci-dir'}.'qdinfo.dat', '/usr/local/share/fci/qdinfo.dat', '/usr/share/fci/qdinfo.dat');
	}
	else {
		@paths = (dirname($0).'\\qdinfo.dat');
	}
		
	foreach(@paths) {
		print " $_\n" if($cfg{verbose});

		# exists and is readable?
		if(-e $_ && -r $_) {
			print "Using qdinfo.dat at: $_\n" if($cfg{verbose});
			$cfg{'qdinfo-file'} = $_;
			$found = 1;
			last;
		}
	}
	
	if(!$found) {
		print "Error: Cannot find qdinfo.dat file Use --qdinfo to set the path.\n";
		exit;
	}
	print "\n" if($cfg{verbose});
	
	return %cfg;
}


##############
#
#     Routine: find_qd_files()
#  Parameters: %cfg
#     Returns: updated %cfg
# Description: check the preset paths for the required programs
# 

sub find_qd_files {
	my %cfg = @_;

	# qd binary
	print "Checking paths for qd binary ... \n" if($cfg{verbose});

	# the qd binary should be in /usr/bin/, /usr/local/bin/ or ~/.fci/ on UNIX systems
	# on Windows the qd binary should be in the same directory as this file, fci-client.pl
	# also try this for backwards compatibility with fci-clients before v1.0
	
	my $found = 0;
	if($^O ne 'MSWin32') {
		my @paths = ('/usr/local/bin/qd', '/usr/bin/qd', $cfg{'fci-dir'}.'qd');
		
		foreach(@paths) {
			$_ .= '.exe' if($^O eq 'cygwin');

			print " $_\n" if($cfg{verbose});

			# exists and is executable?
			if(-e $_ && -x $_) {
				print "Found qd at: $_\n" if($cfg{verbose});
				$cfg{'qd-bin'} = $_;
				$found = 1;
				last;
			}
		}
	}
		
	if(!$found) {
		# in the same dir as this file, fci-client.pl?
		my $qd_bin = dirname($0).$cfg{ds}.$cfg{qd};
		
		print " $qd_bin\n" if($cfg{verbose});
		if(-e $qd_bin && -x $qd_bin) {
			print "Found qd at: $qd_bin\n" if($cfg{verbose});
			$cfg{'qd-bin'} = $qd_bin;
		}
		else {
			print "Error: Cannot find qd binary! Use --qd to set the path.\n";
			exit;
		}
	}
	print "\n" if($cfg{verbose});

	
	# qdinfo.dat file
	print "Checking paths for qdinfo.dat file ... \n" if($cfg{verbose});
	
	# the qdinfo.dat file should be in /usr/share/fci/, /usr/local/share/fci/ or ~/.fci/ on UNIX systems
	# on Windows the qdinfo.dat file should be in the same directory as this file, fci-client.pl
	# also try this for backwards compatibility with fci-clients before v1.0
	
	$found = 0;
	if($^O ne 'MSWin32') {
		my @paths = ('/usr/share/fci/qdinfo.dat', '/usr/local/share/fci/qdinfo.dat', $cfg{'fci-dir'}.'qdinfo.dat');
		
		foreach(@paths) {
			print " $_\n" if($cfg{verbose});

			# exists and is readable?
			if(-e $_ && -r $_) {
				print "Found qdinfo.dat at: $_\n" if($cfg{verbose});
				$cfg{'qdinfo-file'} = $_;
				$found = 1;
				last;
			}
		}
	}
	
	if(!$found) {
		# in the same dir as this file, fci-client.pl?
		my $qdinfo_file = dirname($0).$cfg{ds}.'qdinfo.dat';

		print " $qdinfo_file\n" if($cfg{verbose});
		if(-e $qdinfo_file && -r $qdinfo_file) {
			print "Found qdinfo.dat at: $qdinfo_file\n" if($cfg{verbose});
			$cfg{'qdinfo-file'} = $qdinfo_file;
		}
		else {
			print "Error: Cannot find qdinfo.dat file Use --qdinfo to set the path.\n";
			exit;
		}
	}
	print "\n" if($cfg{verbose});
	
	return %cfg;
}


##############
#
#     Routine: valid_permissions()
#  Parameters: none
#     Returns: 1 if all permissions check out,
#              0 if one or more permissions don't check out
# Description: verify permissions of all files required for operation
# 

sub valid_permissions {
	print "Verifying file permissions and existance:\n" if($cfg{verbose});
		
	# Verify on Windows system or Linux/FreeBSD/OpenBSD 
	if($^O ne 'MSWin32') {
		# Not on windows, assume UNIX filesystem
	
		# required: $cfg{'fci-dir'}
		# requires: write: files need to be updated and/or created there
		
		print " Checking: $cfg{'fci-dir'} ... " if($cfg{verbose});
		if(-w $cfg{'fci-dir'}) {
			print "OK\n" if($cfg{verbose});
		}
		else {
			print "FAILED!\n" if($cfg{verbose});
			print "Error: Cannot write FCI directory! ($cfg{'fci-dir'})\n";
		
			return 0;
		}
		
		# required: qdinfo.dat
		# requires: read : qd binary needs its content
		#	    write: needs to be updated with new version if available
		
		print " Checking: $cfg{'qdinfo-file'} ... " if($cfg{verbose});
		if(-r $cfg{'qdinfo-file'} && -w $cfg{'qdinfo-file'}) {
			print "OK\n" if($cfg{verbose});
		}
		else {
			print "FAILED!\n" if($cfg{verbose});
			print "Error: Cannot read and/or write qdinfo.dat! ($cfg{'qdinfo-file'})\n";
		
			return 0;
		}
		
		# required: qd | qd.exe | qd.darwin
		# requires: write: needs to be updated with new version if available
		#           exec : needs to be executed to parse queue.dat
		
		print " Checking: $cfg{'qd-bin'} ... " if($cfg{verbose});
		if(-w $cfg{'qd-bin'} && -x $cfg{'qd-bin'}) {
			print "OK\n" if($cfg{verbose});
		}
		else {
			print "FAILED!\n" if($cfg{verbose});
			print "Error: Cannot write and/or execute qd binary! ($cfg{'qd-bin'})\n";
		
			return 0;
		}
		
		# required: queue.dat
		# requires: read : qd binary needs to parse it
			
		print " Checking: $cfg{'queue-file'} ... " if($cfg{verbose});
		if(-r $cfg{'queue-file'}) {
			print "OK\n" if($cfg{verbose});
		}
		else {
			print "FAILED!\n" if($cfg{verbose});
			print "Error: Cannot read queue.dat! ($cfg{'queue-file'})\n";
		
			return 0;
		}
	
		# optional: current.xyz
		# requires: read : needs to be sent to server
		
		print " Checking: $cfg{'xyz-file'} ... " if($cfg{verbose});
		if(-e $cfg{'xyz-file'}) {
			if(-r $cfg{'xyz-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				$skip_file{'xyz-file'} = "Unreadable";
				
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			$skip_file{'xyz-file'} = "Nonexistent";
	
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}
		
		# optional: client.cfg
		# requires: read : needs to be sent to server
		
		print " Checking: $cfg{'client-cfg-file'} ... " if($cfg{verbose});
		if(-e $cfg{'client-cfg-file'}) {
			if(-r $cfg{'client-cfg-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				$skip_file{'client-cfg-file'} = "Unreadable";
				
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			$skip_file{'client-cfg-file'} = "Nonexistent";
	
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}
	
		# optional: FAHlog.txt
		# requires: read : needs to be sent to server
		
		print " Checking: $cfg{'fahlog-file'} ... " if($cfg{verbose});
		if(-e $cfg{'fahlog-file'}) {
			if(-r $cfg{'fahlog-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				$skip_file{'fahlog-file'} = "Unreadable";
				
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			$skip_file{'fahlog-file'} = "Nonexistent";
	
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}

		# optional: FAHlog-Prev.txt
		# requires: read : needs to be sent to server

		print " Checking: $cfg{'fahlog-prev-file'} ... " if($cfg{verbose});
		if(-e $cfg{'fahlog-prev-file'}) {
			if(-r $cfg{'fahlog-prev-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				$skip_file{'fahlog-prev-file'} = "Unreadable";
				
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			$skip_file{'fahlog-prev-file'} = "Nonexistent";
	
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}
	
		# optional: unitinfo.txt
		# requires: read : needs to be sent to server
	
		print " Checking: $cfg{'unitinfo-file'} ... " if($cfg{verbose});
		if(-e $cfg{'unitinfo-file'}) {
			if(-r $cfg{'unitinfo-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				$skip_file{'unitinfo-file'} = "Unreadable";
			
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			$skip_file{'unitinfo-file'} = "Nonexistent";
	
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}
	
		# optional: fci-client.conf
		# requires: read : needs to be parsed for config overwrites
		#           write: needs to updated with new config
	
		print " Checking: $cfg{'fci-conf-file'} ... " if($cfg{verbose});
		if(-e $cfg{'fci-conf-file'}) {
			if(-r $cfg{'fci-conf-file'} && -w $cfg{'fci-conf-file'}) {
				print "OK\n" if($cfg{verbose});
			}
			else {
				print "FAILED! (OK, not required)\n" if($cfg{verbose});
			}
		}
		else {
			print "MISSING! (OK, not required)\n" if($cfg{verbose});
		}
	
		return 1;
	}
	else {
		# Running on Windows, check files appropriately
		
		# FIXME! Actually check the files....	
		return 1;
	}
}


##############
#
#     Routine: verify_fci_dir()
#  Parameters: none
#     Returns: 1 on success
#              0 in case of error
# Description: check the existance and content of the fci directory
# 

sub verify_fci_dir {
	print "Verifying fci directory: $cfg{'fci-dir'} ... " if($cfg{verbose});
	
	# Verify on Windows system or Linux/FreeBSD/OpenBSD 
	if($^O ne 'MSWin32') {
		# Not on windows, assume UNIX filesystem
		
		my $qd  = 'qd';
		   $qd .= '.exe' if($^O eq 'cygwin');

		# does it exist at all?
		# If it doesn't exist, create it
		if(-e $cfg{'fci-dir'}) {
			print "Exists.\n" if($cfg{verbose});

			# verify read and write permissions on directory
			if(-r $cfg{'fci-dir'} && -w $cfg{'fci-dir'}) {
				
				if(-e $cfg{'fci-dir'}.$qd) {
					# verify write & execute on qd binary
					if(!-w $cfg{'fci-dir'}.$qd || !-x $cfg{'fci-dir'}.$qd) {
						print "Error: Cannot write and/or execute qd binary! (".$cfg{'fci-dir'}.$qd.")\n";
						return 0;
					}
				}
				else {
					# the file does not exists,
					# try to copy it there
					
					print " qd binary not found!\n";
					print " Copying qd binary: $cfg{'qd-bin'} ... " if($cfg{verbose});
					if(copy $cfg{'qd-bin'}, $cfg{'fci-dir'}.$qd) {
						print "Success.\n" if($cfg{verbose});
	
						# set read, write & execute permissions
						my $mode = 0700;
						chmod $mode, $cfg{'fci-dir'}.$qd;
					}
					else {
						print "FAILED!\n" if($cfg{verbose});
						return 0;
					}
				}
				
				if(-e $cfg{'fci-dir'}.'qdinfo.dat') {
					# verify read & write on qdinfo.dat file
					if(!-r $cfg{'fci-dir'}.'qdinfo.dat' || !-w $cfg{'fci-dir'}.'qdinfo.dat') {
						print "Error: Cannot read and/or write qdinfo.dat file! (".$cfg{'fci-dir'}.'qdinfo.dat'.")\n";
						return 0;
					}
				}
				else {
					# the file does not exists,
					# try to copy it there
					
					print " qdinfo.dat file not found!\n";
					print " Copying qdinfo.dat file: $cfg{'master-qdinfo-file'} ... " if($cfg{verbose});
					if(copy $cfg{'qdinfo-file'}, $cfg{'fci-dir'}.'qdinfo.dat') {
						print "Success.\n" if($cfg{verbose});
					
						# set read & write permissions
						my $mode = 0600;
						chmod $mode, $cfg{'fci-dir'}.'qdinfo.dat';
					}
					else {
						print "FAILED!\n" if($cfg{verbose});
						return 0;
					}
				}
				
				# verify read & write on fci-client.conf if it exists
				if(-e $cfg{'fci-conf-file'}) {
					if(!-r $cfg{'fci-conf-file'} || !-w $cfg{'fci-conf-file'}) {
						print "Error: Cannot write and/or execute fci configuration file! ($cfg{'fci-conf-file'})\n";
						return 0;
					}
				}
				
				return 1;
			}
			else {
				print "Error: Cannot read and/or write fci directory! ($cfg{'fci-dir'})\n";
				return 0;
			}
		}
		else {
			print "Missing.\n" if($cfg{verbose});
	
			print "Creating directory $cfg{'fci-dir'} ... " if($cfg{verbose});
			if(mkdir $cfg{'fci-dir'}, 0700) {
				print "Success.\n" if($cfg{verbose});
	
				# copy required files
				print " Copying qd binary: $cfg{'qd-bin'} ... " if($cfg{verbose});
				if(copy $cfg{'qd-bin'}, $cfg{'fci-dir'}.$qd) {
					print "Success.\n" if($cfg{verbose});

					# set read, write & execute permissions
					my $mode = 0700;
					chmod $mode, $cfg{'fci-dir'}.$qd;
				}
				else {
					print "FAILED!\n" if($cfg{verbose});
					return 0;
				}
				
				print " Copying qdinfo.dat file: $cfg{'qdinfo-file'} ... " if($cfg{verbose});
				if(copy $cfg{'qdinfo-file'}, $cfg{'fci-dir'}.'qdinfo.dat') {
					print "Success.\n" if($cfg{verbose});
				
					# set read & write permissions
					my $mode = 0600;
					chmod $mode, $cfg{'fci-dir'}.'qdinfo.dat';
				}
				else {
					print "FAILED!\n" if($cfg{verbose});
					return 0;
				}
	
				return 1;
			}
			else {
				print "FAILED! ($!)\n" if($cfg{verbose});
				
				return 0;
			}
		}
	}
	else {
		# Running on Windows, check files appropriately
		
		# Check files:
		# - qd.exe	read/write/execute
		# - qdinfo.dat  read/write
		
		if(-e $cfg{'fci-dir'}) {
			print "Exists.\n" if($cfg{verbose});
			
			# qd binary present
			if(-e $cfg{'fci-dir'}.'qd.exe') {
				if(!-w $cfg{'fci-dir'}.'qd.exe' || !-x $cfg{'fci-dir'}.'qd.exe') {
					print "Error: Cannot write and/or execute qd binary! (".$cfg{'fci-dir'}.'qd.exe'.")\n";
					return 0;
				}
			}
			else {
				print "Error: Cannot find qd binary! (".$cfg{'fci-dir'}.'qd.exe'.")\n";
				return 0;
			}

			# qdinfo.dat file present
			if(-e $cfg{'fci-dir'}.'qdinfo.dat') {
				if(!-r $cfg{'fci-dir'}.'qdinfo.dat' || !-w $cfg{'fci-dir'}.'qdinfo.dat') {
					print "Error: Cannot read and/or write qdinfo.dat file! (".$cfg{'fci-dir'}.'qdinfo.dat'.")\n";
					return 0;
				}
			}
			else {
				print "Error: Cannot find qd binary! (".$cfg{'fci-dir'}.'qd.exe'.")\n";
				return 0;
			}
		}
		else {
			print "MISSING!\n" if($cfg{verbose});
			return 0;
		}

		return 1;
	}
}


##############
#
#     Routine: set_fci_dir()
#  Parameters: none
#     Returns: full path to fci directory
# Description: find the full path to the fci directory depending on 
#              which Operating System is running
# 

sub set_fci_dir {
	my $dir  = '/home/fci/';
	
	if($^O eq 'MSWin32') {
		# Windows, should be same directory as this script
		$dir = dirname($0).'\\';
	}
	else {
		# Linux/UNIX is assumed, should be in dotdir users home directory
		$dir = $ENV{HOME}.'/.fci/';
	}
	
	return $dir;
}


##############
#
#     Routine: load_config()
#  Parameters: none
#     Returns: hash of configuration options
# Description: load the configuration options from file, 
#              skip those given via commandline parameters
#

sub load_config {
	my %cfg = @_;

	print "Loading configuration from file: $cfg{'fci-conf-file'}.\n" if($cfg{verbose});

	open(F, $cfg{'fci-conf-file'}) || return 0;	
	while(<F>) {
		if(!/^\s*#/ && /^\s*(.*?)\s*:\s*(.*?)\s*$/) {
			my ($name, $value) = ($1, $2);
			
			if(!$parameter{$name} && $name eq 'url' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'username' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'password' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": ********\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'useragent' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'client-name' && $value =~ /^[0-9a-zA-Z\-\_]+/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'fah-dir' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'fci-dir' && $value !~ /^\s*$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name eq 'no-qd-update' && $value =~ /^[0|1]$/) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(xyz)$/ && $value == 1) { 
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(client-cfg)$/ && $value == 1) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(fahlog)$/ && $value == 1) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(fahlog-prev)$/ && $value == 1) { 
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(myfolding)$/ && $value == 1) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^skip-(unitinfo)$/ && $value == 1) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$skip_file{$1.'-file'} = "Configuration Override";
				$from_conf{$name} = 1;
			}
			elsif(!$parameter{$name} && $name =~ /^verbose$/ && $value == 1) {
				if($cfg{verbose}) {
					print " $name";
					print " " x (17 - length($name));
					print ": $value\n";
				}
				$cfg{$name} = $value;
				$from_conf{$name} = 1;
			}
			else {
				# unknown option
			}
		}
		else {
			# comment, whitespace or other unparsed line
		}
	}
	close F;

	return %cfg;
}


##############
#
#     Routine: save_config()
#  Parameters: none
#     Returns: 1 on success
#              0 in case of an error
# Description: save the configuration options to file
#

sub save_config {
	my %cfg = @_;

	print "Saving configuration to file: $cfg{'fci-conf-file'}.\n" if($cfg{verbose});

	if(!-e $cfg{'fci-conf-file'} || -w $cfg{'fci-conf-file'}) {
		my $data  = "#\n";
		   $data .= "# File: ".basename($cfg{'fci-conf-file'})."\n";
		   $data .= "#\n";
		   $data .= "# $cfg{self} configuration file\n";
		   $data .= "#\n\n";
		
		if($parameter{url} && $cfg{url} && $cfg{url} !~ /^\s*$/) {
			$data .= "# FCI Server URL - Parameter: --url\n";
			$data .= "url: $cfg{url}\n\n";
		}
		if($parameter{username} && $cfg{username} && $cfg{username} !~ /^\s*$/) {
			$data .= "# FCI Username - Parameter: --username\n";
			$data .= "username: $cfg{username}\n\n";
		}
		if($parameter{password} && $cfg{password} && $cfg{password} !~ /^\s*$/) {
			$data .= "# FCI Password - Parameter: --password\n";
			$data .= "password: $cfg{password}\n\n";
		}
		if($parameter{useragent} && $cfg{useragent} && $cfg{useragent} !~ /^\s*$/) {
			$data .= "# HTTP User-Agent - Parameter: --useragent\n";
			$data .= "useragent: $cfg{useragent}\n\n";
		}
		if($parameter{'client-name'} && $cfg{'client-name'} && $cfg{'client-name'} =~ /^[0-9a-zA-Z\-\_]+/) {
			$data .= "# FCI client name - Parameter: --name | --client (depricated)\n";
			$data .= "client-name: $cfg{'client-name'}\n\n";
		}
		if($parameter{'fah-dir'} && $cfg{'fah-dir'} && $cfg{'fah-dir'} !~ /^\s*$/) {
			$data .= "# Folding\@Home client directory - Parameter: --dir\n";
			$data .= "fah-dir: $cfg{'fah-dir'}\n\n";
		}
		if($parameter{'fci-dir'} && $cfg{'fci-dir'} && $cfg{'fci-dir'} !~ /^\s*$/) {
			$data .= "# FCI directory - Parameter: --fci-dir\n";
			$data .= "fci-dir: $cfg{'fci-dir'}\n\n";
		}
		if($parameter{'qd-bin'} && $cfg{'qd-bin'} && $cfg{'qd-bin'} =~ /^[0|1]$/) {
			$data .= "# Path to the qd executable - Parameter: --qd\n";
			$data .= "qd-bin: $cfg{'qd-bin'}\n\n";
		}
		if($parameter{'qdinfo-file'} && $cfg{'qdinfo-file'} && $cfg{'qdinfo-file'} =~ /^[0|1]$/) {
			$data .= "# Path to the qdinfo.dat file - Parameter: --qdinfo\n";
			$data .= "qdinfo-file: $cfg{'qdinfo-file'}\n\n";
		}
		if($parameter{'update-qd'} && $cfg{'update-qd'} && $cfg{'update-qd'} =~ /^[0|1]$/) {
			$data .= "# Update the qd binary - Parameter: --update-qd\n";
			$data .= "update-qd: $cfg{'update-qd'}\n\n";
		}
		if($parameter{'no-qd-update'} && $cfg{'no-qd-update'} && $cfg{'no-qd-update'} =~ /^[0|1]$/) {
			$data .= "# Don't update qd files - Parameter: --no-qd-update\n";
			$data .= "no-qd-update: $cfg{'no-qd-update'}\n\n";
		}
		
		if($parameter{'skip-xyz'} && $skip_file{'xyz-file'} && $skip_file{'xyz-file'} ne '') { 
			$data .= "# Don't upload the file current.xyz - Parameter: --skip-xyz\n";
			$data .= "skip-xyz: 1\n\n";
		}
		if($parameter{'skip-client-cfg'} && $skip_file{'client-cfg-file'} && $skip_file{'client-cfg-file'} ne '') {
			$data .= "# Don't upload the file client.cfg - Parameter: --skip-client-cfg\n";
			$data .= "skip-client-cfg: 1\n\n";
		}
		if($parameter{'skip-unitinfo'} && $skip_file{'unitinfo-file'} && $skip_file{'unitinfo-file'} ne '') {
			$data .= "# Don't upload the file unitinfo.txt - Parameter: --skip-unitinfo\n";
			$data .= "skip-unitinfo: 1\n\n";
		}
		if($parameter{'skip-fahlog'} && $skip_file{'fahlog-file'} && $skip_file{'fahlog-file'} ne '') {
			$data .= "# Don't upload the file FAHlog.txt - Parameter: --skip-fahlog\n";
			$data .= "skip-fahlog: 1\n\n";
		}
		if($parameter{'skip-fahlog-prev'} && $skip_file{'fahlog-prev-file'} && $skip_file{'fahlog-prev-file'} ne '') { 
			$data .= "# Don't upload the file FAHlog-Prev.txt - Parameter: --skip-fahlog-prev\n";
			$data .= "skip-fahlog-prev: 1\n\n";
		}
		if($parameter{'skip-myfolding'} && $skip_file{'myfolding-file'} && $skip_file{'myfolding-file'} ne '') {
			$data .= "# Don't upload the file MyFolding.html - Parameter: --skip-myfolding\n";
			$data .= "skip-myfolding: 1\n\n";
		}

		if($parameter{'verbose'} && $cfg{'verbose'} && $cfg{'verbose'} =~ /^[0|1]$/) {
			$data .= "# Verbose (Debug) Output - Parameter: --verbose\n";
			$data .= "verbose: 1\n\n";
		}

		return save_file($cfg{'fci-conf-file'}, $data, mode => '0600');
	}
	else {
		print "Error: Cannot write configuration file! ($cfg{'fci-conf-file'})\n";
		return 0;
	}
}

##############
#
#     Routine: find_binary()
#  Parameters: default return value
#     Returns: path to the specified binary
# Description: find the path to the specified binary or return the default value;
# 

sub find_binary {
	my $name = shift;
	my $file = shift;

	if(chomp(my $path = `which $name 2> /dev/null`)) {
		return $path;
	}
	else {
		my @dir = qw|/bin/ /usr/bin/ /usr/local/bin/ /sbin/ /usr/sbin/ /usr/local/sbin/|;

		foreach(@dir) {
			my $bin  = $_;
			   $bin .= '/' if(substr($path, -1, 1) ne '/');
			   $bin .= $name;

			return $bin if(-x $bin);
		}
	}

	return $file;
}


##############
#
#     Routine: get_lsb_release_data()
#  Parameters: lsb_release path
#     Returns: hash with LSB info
# Description: get the Linux Standard Base info for this host
# 

sub get_lsb_release_data {
	my $file = shift;
	
	my %info;
	if(-x $file) {
		#$ lsb_release -a
		#LSB Version:    n/a
		#Distributor ID: Debian
		#Description:    Debian GNU/Linux
		#Release:        3.1
		#Codename:       sarge
	
		if(my $data = `$file --all 2>&1`) {
			if($data =~ /LSB Version:\t(.*)/) {
				$info{lsb_version} = $1;
			}
			if($data =~ /Distributor ID:\t(.*)/) {
				$info{distrib_id} = $1;
			}
			if($data =~ /Description:\t(.*)/) {
				$info{distrib_description} = $1;
			}
			if($data =~ /Release:\t(.*)/) {
				$info{distrib_release} = $1;
			}
			if($data =~ /Codename:\t(.*)/) {
				$info{distrib_codename} = $1;
			}
		}
	}

	return %info;
}


##############
#
#     Routine: print_usage()
#  Parameters: none
#     Returns: nothing
# Description: print usage information: version, parameters, etc
# 

sub print_usage {
	my $verbose         = $cfg{'verbose'}                == 1 ? 'on'   : 'off';
	my $update_qd       = $cfg{'update-qd'}              == 1 ? 'true' : 'false';
	my $no_qd_update    = $cfg{'no-qd-update'}           == 1 ? 'true' : 'false';
	my $save_config     = $cfg{'save-config'}            == 1 ? 'true' : 'false';
	my $xyz_file        = 'false';
	my $clientcfg_file  = 'false';
	my $fahlog_file     = 'false';
	my $fahlogprev_file = 'false';
	my $myfolding_file  = 'false';
	my $unitinfo_file   = 'false';
	
	if($skip_file{'xyz-file'})         { $xyz_file        = 'true'; }
	if($skip_file{'client-cfg-file'})  { $clientcfg_file  = 'true'; }
	if($skip_file{'fahlog-file'})      { $fahlog_file     = 'true'; }
	if($skip_file{'fahlog-prev-file'}) { $fahlogprev_file = 'true'; }
	if($skip_file{'myfolding-file'})   { $myfolding_file  = 'true'; }
	if($skip_file{'unitinfo-file'})    { $unitinfo_file   = 'true'; }
	
	my $qd_version    = '';
	chomp($qd_version = `"$cfg{'qd-bin'}" -v`) if(-x $cfg{'qd-bin'});
	
	print "$cfg{self} v$cfg{version}";
	print " with qd [$qd_version]" if($qd_version);
	print "\n\n";
	print "Usage: ".basename($0)." [OPTIONS]\n\n";
	print "Options:\n";
	print "--url <url>           URL to POST the data to               ($cfg{url})\n";
	print "--username <string>   Server authentication username        ($cfg{username})\n";
	print "--password <string>   Server authentication password        ($cfg{password})\n";
	print "--useragent <name>    User-Agent string                     ($cfg{useragent})\n";
	print "\n";
	print "--name <name>         Client name                           ($cfg{'client-name'})\n";
	print "--dir <directory>     Folding\@Home client directory         ($cfg{'fah-dir'})\n";
	print "--skip-xyz            Don't upload the file current.xyz     ($xyz_file)\n";
	print "--skip-client-cfg     Don't upload the file client.cfg      ($clientcfg_file)\n";
	print "--skip-unitinfo       Don't upload the file unitinfo.txt    ($unitinfo_file)\n";
	print "--skip-fahlog         Don't upload the file FAHlog.txt      ($fahlog_file)\n";
	print "--skip-fahlog-prev    Don't upload the file FAHlog-Prev.txt ($fahlogprev_file)\n";
	print "--skip-myfolding      Don't upload the file MyFolding.html  ($myfolding_file)\n";
	print "\n";
	print "--fci-dir <directory> FCI directory                         ($cfg{'fci-dir'})\n";
	print "--qd <path>           Path to the qd executable             ($cfg{'qd-bin'})\n";
	print "--qdinfo <path>       Path to the qdinfo.dat file           ($cfg{'qdinfo-file'})\n";
	print "--update-qd           Update the qd binary                  ($update_qd)\n";
	print "--no-qd-update        Don't update any qd files             ($no_qd_update)\n";
	print "--save-config         Save the configuration to file        ($save_config)\n";
	print "--config-file <path>  Path to the configuration file        ($cfg{'fci-conf-file'})\n";
	print "\n";
	print "--verbose             Verbose (debug) output                ($verbose)\n";
	print "--help                Print this information\n";
}
